import warnings
from dataclasses import dataclass
from datetime import datetime
from typing import Any, List, Literal, Optional, cast

import empyrical as ep
import numpy as np
import pandas as pd
from plotnine import (
    aes,
    element_blank,
    element_text,
    facet_grid,
    facet_wrap,
    geom_col,
    geom_line,
    geom_point,
    geom_rect,
    geom_text,
    ggplot,
    position_dodge,
    scale_x_datetime,
    theme,
    ylab,
    ylim,
)

from portfolio_plan._properties import Frequency, Weights
from portfolio_plan.visualisation import (
    THEME_ROSE_PINE_BASE_SIZE,
    scale_alphabet_fill_discrete,
    scale_brewer_fill_discrete,
    scale_rose_pine_discrete,
    scale_rose_pine_fill_discrete,
    theme_rose_pine,
)


def validate_dataframe(data: pd.DataFrame, check_frequency: bool = True):
    assert isinstance(data, pd.DataFrame), "Data must be a pandas DataFrame"
    assert not data.empty, "DataFrame should not be empty"
    columns = data.columns.tolist()
    expected_columns = ["open", "close", "high", "low"]
    assert set(columns) == set(expected_columns), (
        f"DataFrame must have exactly columns {expected_columns}. Got {columns}"
    )

    assert isinstance(data.index, pd.MultiIndex), "Index must be a MultiIndex"
    expected_index_names = ["symbol", "date"]
    index_names = data.index.names
    assert index_names == expected_index_names, (
        f"Index names must be {expected_index_names} got {index_names}"
    )

    row_has_nan = data.isna().any(axis=1)
    row_all_nan = data.isna().all(axis=1)
    rows_with_partial_nan = row_has_nan & ~row_all_nan

    if rows_with_partial_nan.any():
        first_bad_row = rows_with_partial_nan[rows_with_partial_nan].index[0]
        symbol, date = first_bad_row
        values = data.loc[first_bad_row]

        raise AssertionError(
            f"Invalid NaN pattern found for symbol {symbol} at date {date}. "
            f"If a row contains NaN, all values in that row must be NaN. "
            f"Got values: {values.to_dict()}"
        )
    for symbol, df in data.groupby(level="symbol"):
        date_index: pd.DatetimeIndex = df.index.get_level_values("date")  # type: ignore

        assert (
            date_index.is_monotonic_increasing
        ), f"""Dates must be sorted for symbol {symbol}.
            Are dates unique ? Got {date_index}
            """
        if check_frequency:
            _validate_frequency(date_index)


def _validate_frequency(index: pd.DatetimeIndex, symbol: Optional[str] = None):
    if index.shape[0] > 1:
        inferred_freq = pd.infer_freq(index)
        if symbol is not None:
            assert inferred_freq is not None, (
                f"Date index must have an inferrable frequency for symbol {symbol}. Found freq=None"
            )
        else:
            assert inferred_freq is not None, (
                "Date index must have an inferrable frequency for current symbol. Found freq=None"
            )

        return
    # else, the index is a singleton


def _returns_from_prices(prices: pd.DataFrame):
    """
    Calculate the returns

    Parameters
    ----------
    prices : pd.DataFrame
        OHLC price data

    Returns
    -------
    pd.DataFrame
        Returns derived from prices
    """
    # TODO: check that NaS are equivalent in all OHLC dimension
    index = prices.index

    mask = prices.isna().all(axis=1).to_numpy()
    out = np.full(prices.shape, np.nan, dtype=np.float64)
    pct_change = prices[~mask].pct_change().to_numpy()
    out[~mask] = pct_change
    returns = pd.DataFrame(out, index=index, columns=prices.columns)
    return returns


def returns_from_prices(
    data: pd.DataFrame, check_frequency: bool = True
) -> pd.DataFrame:
    """
    Parameters
    ==========
    data: pd.DataFrame
        Take an input DataFrame, transform it.
    check_frequency: bool
        See :py:meth:`validate_dataframe`
        Defaults to True
    """
    validate_dataframe(data, check_frequency=check_frequency)
    symbols = data.index.get_level_values("symbol")
    first_symbol = symbols[0]
    freq = data.loc[first_symbol].index.freq
    prices = data.unstack(level="symbol")  # Creates multiindex column
    prices = cast(pd.DataFrame, prices)
    returns = _returns_from_prices(prices)
    if freq is not None:
        returns = returns.asfreq(freq)
    stacked_returns = returns.stack(dropna=False).reorder_levels([1, 0])
    stacked_returns.columns.name = "Returns"

    return stacked_returns


# TODO: test with singleton
def cumulative_from_returns(
    data: pd.DataFrame, starting_value: float = 1.0, keepna: bool = True
) -> pd.DataFrame:
    # TODO: Implement validation of data
    assert isinstance(data, pd.DataFrame)
    usymbols = data.index.get_level_values("symbol").unique()
    newreturns: List[pd.DataFrame] = []
    for symbol in usymbols:
        returns = data.loc[symbol, :]
        if keepna:
            not_a_nan_row = ~returns.isna().any(axis=1)
            if not_a_nan_row[0] == np.False_:
                not_a_nan_row[0] = np.True_
            out = np.empty(returns.shape, dtype=np.float64)
            out[:] = np.nan
            returns_without_nan = returns.loc[not_a_nan_row, :]

            # NOTE: Probably the element causing the discrepency
            out[not_a_nan_row] = ep.cum_returns(
                returns_without_nan.values,
                starting_value=starting_value,
                out=out[not_a_nan_row, :],
            )
            cumulative_returns = pd.DataFrame(
                out, columns=returns.columns, index=returns.index
            )
        else:
            cumulative_returns = ep.cum_returns(returns, starting_value=starting_value)
        cumulative_returns.name = symbol
        newreturns.append(cumulative_returns)

    newreturns_data: pd.DataFrame = pd.concat(  # type: ignore
        newreturns,
        axis=0,
        keys=[df.name for df in newreturns],
        names=["symbol", "date"],
    )
    newreturns_data.columns.name = "CumulativeReturns"
    return newreturns_data


def _find_gap_lines(data: pd.Series):
    """
    Find gaps (even multiple missing points) and build dashed lines to connect
    valid data.
    The function connects the last valid point before a gap to the
    first valid point after the gap.
    """
    name = data.name

    if data.empty:
        raise ValueError("Empty Series, cannot find gaps in empty data.")

    data = data.sort_index()
    records = []
    # TODO: ignore blocks of NAN which start at index 0 (there's no before)
    # & Ignore blocks of NaN that end at index -1 (there's no after)
    for symbol, group in data.groupby(level="symbol"):
        series = group.droplevel("symbol")

        is_na = series.isna()
        if not is_na.any():
            continue

        gap_start_idx = None
        last_valid_value = None
        last_valid_date = None

        for idx, (date, value) in enumerate(series.items()):
            if pd.isna(value):
                if idx == 0:
                    continue
                # If we find a NaN and we haven't started a gap yet
                if gap_start_idx is None and last_valid_value is not None:
                    gap_start_idx = idx
            else:
                # If we have a valid value and we were in a gap
                if gap_start_idx is not None:
                    # Connect the last valid point before the gap to this point
                    records.append(
                        {
                            "symbol": symbol,
                            "date_start": last_valid_date,
                            "value_start": last_valid_value,
                            "date_end": date,
                            "value_end": value,
                        }
                    )
                    gap_start_idx = None

                # Update the last valid point
                last_valid_value = value
                last_valid_date = date

    if not records:
        return pd.DataFrame()

    gap_lines = pd.DataFrame(records)

    # Flatten into long format
    gap_lines_long = pd.DataFrame(
        {
            "symbol": np.ravel(
                np.column_stack(
                    [gap_lines["symbol"].to_numpy(), gap_lines["symbol"].to_numpy()]
                )
            ),
            "date": np.ravel(
                np.column_stack(
                    [
                        gap_lines["date_start"].to_numpy(),
                        gap_lines["date_end"].to_numpy(),
                    ]
                )
            ),
            name: np.ravel(
                np.column_stack(
                    [
                        gap_lines["value_start"].to_numpy(),
                        gap_lines["value_end"].to_numpy(),
                    ]
                )
            ),
            # Add the group column back to ensure proper line segments
            "group": np.repeat(np.arange(len(gap_lines)), 2),
        }
    )

    return gap_lines_long


@dataclass
class Period:
    """"""

    """The start of the period"""
    start: datetime
    """The end of the period"""
    end: datetime
    """The frequency of data within the period"""
    frequency: Frequency | None


def _geom_lines(
    data: pd.Series, periods: List[Period] | None = None
) -> List[geom_line | geom_point]:
    name = data.name

    if data.empty:
        return []

    data_subsets = []
    gap_lines_list = []

    if periods:
        dates = data.index.get_level_values("date")
        for period in periods:
            subset = data.loc[(dates >= period.start) & (dates <= period.end)]
            gap_lines_list.append(_find_gap_lines(subset))
            data_subsets.append(subset)
    else:
        # NOTE: Append a single item
        data_subsets.append(data)
        gap_lines_list.append(_find_gap_lines(data))

    geoms: List[geom_line | geom_point] = []
    for gap_lines_long, subset in zip(gap_lines_list, data_subsets):
        geoms.append(
            # Straight line
            geom_line(
                data=subset.reset_index(),
                mapping=aes(x="date", y=name, group="symbol", color="symbol"),
                # TODO: IMPLEMENT COLOR MAPPING FUNCTION
            )
        )

        if not gap_lines_long.empty:
            geoms.append(
                # Dashed line
                geom_line(
                    data=gap_lines_long,
                    mapping=aes(x="date", y=name, group="group", color="symbol"),
                    linetype="dashed",
                )
            )
        group_sizes = subset.groupby(level="symbol").size()
        singleton_groups = group_sizes[group_sizes == 1].index
        singleton_df = subset.loc[singleton_groups, :]
        if not singleton_df.empty:
            geoms.append(
                geom_point(
                    data=singleton_df.reset_index(),
                    mapping=aes(x="date", y=name, group="symbol", color="symbol"),
                )
            )

    return geoms


def _scale_date(dates: pd.DatetimeIndex, periods: List[Period] | None = None):
    if periods:
        assert all(periods[i] != periods[i - 1] for i in range(1, len(periods)))
        freq = periods[0].frequency
    else:
        try:
            inferred = pd.infer_freq(dates)
            if inferred is None:
                raise ValueError("Failed to infer frequency")
            freq = Frequency.from_pandas(inferred)
        except ValueError:
            freq = None

    breaks = "1 year"
    labels = "%Y"

    if freq is not None:
        match freq:
            case Frequency.MINUTE:
                breaks = "1 minute"
                labels = "%H:%M"
            case Frequency.DAILY:
                breaks = "1 day"
                labels = "%Y-%m-%d"
            case Frequency.WEEKLY:
                breaks = "1 week"
                labels = "%Y-%m-%d"
            case Frequency.MONTHLY:
                breaks = "1 month"
                labels = "%Y-%m"
            case _:
                breaks = "1 year"
                labels = "%Y"
    else:
        warnings.warn(
            f"Failed to infer frequency, defaulting to x scale with date_breaks={breaks} and date_labels={labels}"
        )

    return scale_x_datetime(date_breaks=breaks, date_labels=labels)


def plot_lines(
    data: pd.Series,
    scale_x: bool = True,
    periods: List[Period] | None = None,
    variant: Literal["main", "moon", "dawn"] = "dawn",
) -> ggplot:
    """
    Parameters
    ----------
    data: pd.Series
        Data in long form, with multiindex, first level "symbol", second level "date".
    scale_x: bool
        Whether to include a date scale on the x axis
        For data which has uneven frequency, it is recommend to set to False and define
        the scale manually.
        Defaults to True
    periods: List[Period] | None
        If a list of Period is provided, transparent rectangle will be drawn on
        each period.start, period.end
        Defaults to None
    variant: Literal["main", "moon", "dawn"]
        Refer to :py:class:`portfolio_plan.visualisation.theme_rose_pine`
    """
    base_plot = ggplot() + scale_rose_pine_discrete(variant=variant)
    for geom in _geom_lines(data, periods):
        base_plot += geom
    if scale_x:
        symbols = data.index.get_level_values("symbol")
        first_symbol = symbols[0]

        subset = data.loc[first_symbol]
        dates: pd.DatetimeIndex = subset.index.get_level_values("date")  # type: ignore
        base_plot = base_plot + _scale_date(dates, periods)
    if periods is not None:
        periods_data = pd.DataFrame.from_records(
            {
                "start_date": period.start,
                "end_date": period.end,
                "max_value": data.max(),
                "min_value": data.min(),
            }
            for period in periods
        )
        base_plot = base_plot + geom_rect(
            data=periods_data,
            mapping=aes(
                xmin="start_date",
                xmax="end_date",
                ymin="min_value",
                ymax="max_value",
            ),
            alpha=0.2,
        )
    return (
        base_plot
        + theme_rose_pine(variant=variant)
        + theme(axis_text_x=element_text(angle=90))
    )


def plot_prices(data: pd.Series, scale_x: bool = True) -> ggplot:
    """
    A wrapper for :py:func:`plot_lines`
    """
    return plot_lines(data, scale_x)


def plot_returns(data: pd.Series, scale_x: bool = True) -> ggplot:
    """
    A wrapper for :py:func:`plot_lines`, adds a facet layer.
    Each symbol is facetted in order to improve readability of the chart.
    """
    return plot_lines(data) + facet_grid("symbol ~ .")


def plot_cumulative_returns(
    data: pd.Series, scale_x: bool = True, periods=List[Period] | None
) -> ggplot:
    """
    A wrapper for :py:func:`plot_lines`
    """
    return plot_lines(data, scale_x, periods=periods)


def add_singleton_level(data: pd.DataFrame, value: Any) -> None:
    """
    Warning
    -------
    Mutates the original DataFrame
    """
    new_level = [value] * data.shape[0]
    data.index = pd.MultiIndex.from_arrays(
        [new_level, data.index],
        names=["symbol", data.index.name or "date"],
    )


def _create_str_stack(title: str, values: List[str]):
    max_length = max(len(title), *(len(value) for value in values))
    line = "-" * (max_length + 4)

    header = f"{line}\n| {title.ljust(max_length)} |\n{line}"

    content = ""
    for value in values:
        content += f"\n| {value.ljust(max_length)} |\n{line}"

    return header + content


def _geom_text(data: pd.Series, text: str) -> geom_text:
    name = data.name
    labels = np.empty(data.shape[0], dtype="object")
    labels.fill(pd.NA)
    labels[-1] = text
    dtype = pd.StringDtype()
    data = data.reset_index(drop=False)  # type: ignore
    data["label"] = pd.Series(labels, dtype=dtype)

    return geom_text(data=data, mapping=aes(x="date", y=name, label="label"))


def plot_allocations(
    weights: List[Weights] | Weights,
    periods: List[Period] | None = None,
    variant: Literal["main", "moon", "dawn"] = "dawn",
    base_size: int = THEME_ROSE_PINE_BASE_SIZE,
) -> ggplot:
    """
    Plot allocations over multiple periods.

    Parameters
    ----------
    weights : List[Weights] | Weights
        A list of weight dictionaries, where each dictionary represents
        the allocation of weights for a specific period, or a single Weights object.
    periods : List[Period] | None
        A list of Period objects corresponding to the weights, or None for a single allocation.
    variant : Literal["main", "moon", "dawn"]
        The color variant to use for the plot.

    Returns
    -------
    ggplot
        A ggplot object representing the allocations.
    """
    if not isinstance(weights, (list, Weights)):
        raise TypeError("weights must be a List[Weights] or a single Weights object.")
    if periods is not None and not isinstance(periods, list):
        raise TypeError("periods must be a List[Period] or None.")

    if periods is None:
        if not isinstance(weights, Weights):
            raise ValueError(
                "When periods=None, weights must be a single Weights object."
            )
        if weights.empty:
            raise ValueError("Weights can not be empty")
        data = [
            {"symbol": symbol, "weight": weight} for symbol, weight in weights.items()
        ]
        df = pd.DataFrame(data)
    else:
        if len(weights) != len(periods):
            raise ValueError("The number of weights must match the number of periods.")
        data = []
        for w, period in zip(weights, periods):
            if w.empty:
                raise ValueError("Weights can not be empty")
            period_label = (
                f"{period.start.strftime('%b %Y')} - {period.end.strftime('%b %Y')}"
            )
            for symbol, weight in w.items():
                data.append(
                    {
                        "symbol": symbol,
                        "weight": weight,
                        "period_label": period_label,
                    }
                )
        df = pd.DataFrame(data)

    base = (
        ggplot(df, aes(x="symbol", y="weight", fill="symbol"))
        + geom_col()
        + ylim(-0.0, 1.02)
        + theme_rose_pine(variant=variant, base_size=base_size)
        + ylab("Allocation")
    )
    n_symbols = df["symbol"].unique().shape[0]
    if n_symbols <= 6:
        base += scale_rose_pine_fill_discrete(variant=variant)
    elif n_symbols <= 12:
        base += scale_brewer_fill_discrete()
    else:
        base += scale_alphabet_fill_discrete()

    if periods is not None:
        base += facet_wrap("~period_label", scales="free_y")

    return base


def plot_allocations2(
    weights: List[Weights] | Weights,
    periods: List[Period] | None = None,
    variant: Literal["main", "moon", "dawn"] = "dawn",
    base_size: int = THEME_ROSE_PINE_BASE_SIZE,
    text_size: float = 0.8 * THEME_ROSE_PINE_BASE_SIZE,
    text_angle: float = 45,
) -> ggplot:
    """
    Plot allocations over multiple periods.

    Parameters
    ----------
    weights : List[Weights] | Weights
        A list of weight dictionaries, where each dictionary represents
        the allocation of weights for a specific period, or a single Weights object.
    periods : List[Period] | None
        A list of Period objects corresponding to the weights, or None for a single allocation.
    variant : Literal["main", "moon", "dawn"]
        The color variant to use for the plot.

    Returns
    -------
    ggplot
        A ggplot object representing the allocations.
    """
    if not isinstance(weights, (list, Weights)):
        raise TypeError("weights must be a List[Weights] or a single Weights object.")
    if periods is not None and not isinstance(periods, list):
        raise TypeError("periods must be a List[Period] or None.")

    if periods is None:
        if not isinstance(weights, Weights):
            raise ValueError(
                "When periods=None, weights must be a single Weights object."
            )
        data = [
            {"symbol": symbol, "weight": weight} for symbol, weight in weights.items()
        ]
        df = pd.DataFrame(data)
        df["symbol"] = pd.Categorical(df["symbol"], ordered=True)
    else:
        if len(weights) != len(periods):
            raise ValueError("The number of weights must match the number of periods.")
        data = [
            {
                "symbol": symbol,
                "weight": weight,
                "start_date": period.start,
                "end_date": period.end,
                "period_label": f"{period.start.strftime('%b %Y')} - {period.end.strftime('%b %Y')}",
            }
            for w, period in zip(weights, periods)
            for symbol, weight in w.items()
        ]
        df = pd.DataFrame(data)
        df["symbol"] = pd.Categorical(df["symbol"], ordered=True)
        df["period_label"] = pd.Categorical(df["period_label"], ordered=True)
    dodge_text = position_dodge(width=0.9)
    base = (
        ggplot(df, aes(x="period_label", y="weight", fill="symbol"))
        + geom_col(stat="identity", position="dodge", show_legend=False)
        + geom_text(
            aes(y=-0.01, label="symbol"),  # , color="symbol"),
            position=dodge_text,
            size=text_size,
            angle=text_angle,
            va="top",
        )
        + ylim(-0.08, 1.02)
        + theme_rose_pine(variant=variant, base_size=base_size)
        + theme(axis_title_x=element_blank())
        + ylab("Allocation")
    )
    n_symbols = df["symbol"].unique().shape[0]
    if n_symbols <= 6:
        base += scale_rose_pine_fill_discrete(variant=variant)
    elif n_symbols <= 12:
        base += scale_brewer_fill_discrete()
    else:
        base += scale_alphabet_fill_discrete()

    return base


def plot_allocation(
    weights: Weights,
    variant: Literal["main", "moon", "dawn"] = "dawn",
    base_size: int = THEME_ROSE_PINE_BASE_SIZE,
) -> ggplot:
    """
    Plot a single allocation.

    Parameters
    ----------
    weights : Weights
        A dictionary representing the allocation of weights.
    variant : Literal["main", "moon", "dawn"]
        The color variant to use for the plot.

    Returns
    -------
    ggplot
        A ggplot object representing the allocation.
    """
    return plot_allocations(
        weights=weights, periods=None, variant=variant, base_size=base_size
    )



import json
from abc import ABC, abstractmethod
from datetime import datetime
from io import BytesIO
from os import PathLike
from pathlib import Path
from typing import IO, Any, List, cast

import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
import yfinance as yf
from pyarrow.lib import Schema as ArrowSchema

from portfolio_plan._properties import Frequency
from portfolio_plan.financialseries import Prices

EXPECTED_SCHEMA: ArrowSchema = pa.schema(
    cast(
        List[pa.Field],
        [
            pa.field("symbol", pa.string()),
            pa.field("date", pa.timestamp(unit="ns")),
            pa.field("open", pa.float64()),
            pa.field("high", pa.float64()),
            pa.field("low", pa.float64()),
            pa.field("close", pa.float64()),
        ],
    )
)


def read_schema(source: Path | IO[Any]) -> ArrowSchema:
    file = pq.ParquetFile(source)
    return file.schema_arrow


def weakly_equivalent(schema: pa.Schema, expected_schema: pa.Schema) -> bool:
    """
    Checks if two Arrow schemas are weakly equivalent.

    Weak equivalence means:
    - Same set of field names
    - Same data types for corresponding field names
    - Ignores field order and nullability

    Parameters
    ==========
        schema: pa.Schema
        expected_schema: pa.Schema
            The schema to compare against.

    Returns
    =======
        True if schemas are weakly equivalent, False otherwise.
    """

    def normalize_fields(s):
        return {f.name: f.type for f in s}

    return normalize_fields(schema) == normalize_fields(expected_schema)


# TODO: Include weak equality
def validate_schema(schema: ArrowSchema):
    if not weakly_equivalent(schema, EXPECTED_SCHEMA):
        raise ValueError(
            f"""Schema does not match the expected schema.
            Expected:
            {str(EXPECTED_SCHEMA)}

            got {str(schema)}
            """
        )


def has_index(schema: ArrowSchema):
    raw = schema.metadata.get(b"pandas") if schema.metadata else None
    if raw is not None:
        metadata = json.loads(raw)
        index_columns = metadata.get("index_columns")
        if index_columns is not None:
            if not isinstance(index_columns, list):
                return False
            for el in index_columns:
                if not isinstance(el, str):
                    return False
            if set(index_columns) == {"symbol", "date"}:
                return True
    return False


class Resource(ABC):
    """ """

    @abstractmethod
    def fetch(
        self,
        symbols: List[str],
        start_date: datetime,
        end_date: datetime,
        name: str,
        **kwargs,
    ) -> Prices:
        pass


class File(Resource):
    def __init__(self, frequency: Frequency, path: PathLike):
        """
        Parameters
        ----------
            path: PathLike
                Path to a parquet file
            frequency:
                Target frequency
        """
        self._path = path
        self._frequency = frequency

    def fetch(
        self,
        symbols: List[str],
        start_date: datetime,
        end_date: datetime,
        name: str,
        **kwargs,
    ) -> Prices:
        """
        Parameters
        ----------
        Raises
        ------
        ValidationError
        IOError
        """
        path = self._path
        schema = read_schema(Path(path))
        validate_schema(schema)

        df = pd.read_parquet(path)
        if not has_index(schema):
            df = df.set_index(["symbol", "date"])
        symbols_idx = df.index.get_level_values("symbol")

        subset = df.loc[symbols_idx.isin(symbols)]
        dates = subset.index.get_level_values("date")

        subset = subset.loc[(dates >= start_date) & (dates <= end_date)]
        prices = Prices(subset, name=name).resample_period(self._frequency)
        return prices


class YFinance(Resource):
    """
    Fetch and process price data from Yahoo! Finance.

    .. warning::
        **DISCLAIMER:**

        * Yahoo!, Y!Finance, and Yahoo! finance are registered trademarks of Yahoo, Inc.
        * This implementation uses the yfinance package to access Yahoo! Finance data.
        * This code is not affiliated, endorsed, or vetted by Yahoo, Inc.
        * Users are responsible for ensuring their usage complies with Yahoo Finance's Terms of Service.
        * `portfolio-plan` features that interface with `yfinance` are intended for **personal, non-commercial use**

        Relevant Yahoo Terms of Service:

        * API Terms: https://policies.yahoo.com/us/en/yahoo/terms/product-atos/apiforydn/index.htm
        * General Terms: https://legal.yahoo.com/us/en/yahoo/terms/otos/index.html
        * Overall Terms: https://policies.yahoo.com/us/en/yahoo/terms/index.htm
    """

    def __init__(self, frequency: Frequency):
        self._frequency = frequency

    def fetch(
        self,
        symbols: List[str],
        start_date: datetime,
        end_date: datetime,
        name: str,
        **kwargs,
    ) -> Prices:
        """
        Parameters
        ----------
        symbols: List[str]
            List of symbols to download
        start_date: datetime
        end_date: datetime
        name: str

        Raises
        ------
        NetworkError
        IOError
        """
        history = yf.download(
            symbols,
            start=start_date,
            end=end_date,
            interval=self._frequency.to_yf(),
            progress=False,
            auto_adjust=True,
        )
        # NOTE: As freq will add NaN to holidays for example
        history = history.asfreq("B")

        history = history.stack(dropna=False).reorder_levels([1, 0])
        history = history.drop(columns=["Volume"])
        history.columns = history.columns.str.lower()
        history.index.names = ["symbol", "date"]
        return Prices(history, name=name)


class Memory(Resource):
    def __init__(self, data: BytesIO, frequency: Frequency):
        self._data = data
        self._frequency = frequency

    def fetch(
        self,
        symbols: List[str],
        start_date: datetime,
        end_date: datetime,
        name: str,
        **kwargs,
    ):
        # TODO: validate schema from buffer object
        schema = read_schema(self._data)
        validate_schema(schema)

        df = pd.read_parquet(self._data, engine="pyarrow")
        if not has_index(schema):
            df = df.set_index(["symbol", "date"])
        symbols_idx = df.index.get_level_values("symbol")

        subset = df.loc[symbols_idx.isin(symbols)]
        dates = subset.index.get_level_values("date")

        subset = subset.loc[(dates >= start_date) & (dates <= end_date)]
        prices = Prices(subset, name=name).resample_period(self._frequency)
        return prices



from datetime import date
from typing import Optional

import numpy as np
import pandas as pd


def generate_stock_series(
    symbols: list[str],
    start_date: date,
    end_date: date,
    initial_price: float = 100.0,
    volatility: float = 5.0,
    seed: Optional[int] = 1,
) -> pd.Series:
    """
    Generate financial time series data using a Wiener process:

    For t ∈ [0,1], the discrete Wiener process is defined as:
        W_n(t) = (1/√n) ∑_{1≤k≤⌊nt⌋} ξ_k

    The price process is modeled as:
        P(t) = P₀ + σ·W_n(t)

    where:
        P₀ = initial_price (base price)
        σ = volatility (scaling factor)

    Example
    -------

    >>> from datetime import date
    >>> generate_stock_series(
            ["a", "b"],
            start_date=date(2020, 1, 1),
            end_date=date(2020, 12, 31)
        )

    """
    np.random.seed(seed)
    n_symbols = len(symbols)

    business_days = pd.date_range(start=start_date, end=end_date, freq="B")
    n_days = len(business_days)

    n = 1000

    xi = np.random.normal(0, 1, size=(n_symbols, n))

    price_matrix = np.zeros((n_symbols, n_days))

    for i in range(n_days):
        t = (i + 1) / n_days

        nt_floor = int(np.floor(n * t))

        for j in range(n_symbols):
            price_matrix[j, i] = (1 / np.sqrt(n)) * np.sum(xi[j, :nt_floor])

    price_matrix = initial_price + volatility * price_matrix

    symbol_idx = np.repeat(np.arange(n_symbols), n_days)
    date_idx = np.tile(np.arange(n_days), n_symbols)

    index = pd.MultiIndex.from_arrays(
        [np.array(symbols)[symbol_idx], business_days[date_idx]],
        names=["symbol", "date"],
    )

    result = pd.Series(
        price_matrix.flatten("C"),
        index=index,
    )

    return result


def generate_stock_data(
    symbols: list[str],
    start_date: date,
    end_date: date,
    open_initial_price: float = 100.0,
    close_initial_price: float = 200.0,
    low_initial_price: float = 300.0,
    high_initial_price: float = 400.0,
    volatility: float = 5.0,
    seed: Optional[int] = 1,
):
    open = generate_stock_series(
        symbols, start_date, end_date, open_initial_price, volatility, seed
    )
    open.name = "open"
    close = generate_stock_series(
        symbols, start_date, end_date, close_initial_price, volatility, seed
    )
    close.name = "close"
    low = generate_stock_series(
        symbols, start_date, end_date, low_initial_price, volatility, seed
    )
    low.name = "low"

    high = generate_stock_series(
        symbols, start_date, end_date, high_initial_price, volatility, seed
    )
    high.name = "high"
    return pd.concat([open, close, low, high], axis=1)



""" """

from datetime import datetime
from enum import Enum
from functools import lru_cache
from typing import List, Literal

import pandas as pd
from plotnine import ggplot

from portfolio_plan import const, resource
from portfolio_plan._properties import Frequency, Weights
from portfolio_plan.errors import PeriodOverlapError
from portfolio_plan.financialseries import CumulativeReturns, Prices, Returns, vstack
from portfolio_plan.prettyprint import ascii_graph
from portfolio_plan.ratios import InformationRatio
from portfolio_plan.utils import (
    Period,
    _create_str_stack,
    plot_allocation,
    plot_allocations,
    plot_allocations2,
    returns_from_prices,
)
from portfolio_plan.visualisation import THEME_ROSE_PINE_BASE_SIZE


class Asset:
    def __init__(self, symbol: str, resource: resource.Resource | None = None):
        """
        Example
        -------

        >>> from portfolio_plan import resource
        >>> asset1 = Asset("A", resource = resource.File("prices.parquet"))
        >>> asset2 = Asset("B", resource = resource.YFinance())
        """
        self._symbol = symbol
        self._resource = resource

    def __str__(self):
        return f"{self.__class__.__name__}(symbol = {self._symbol}, resource={str(self._resource)})"

    def fetch(self, start_date: datetime, end_date: datetime, **kwargs) -> Prices:
        if self._resource is not None:
            prices = self._resource.fetch(
                [self._symbol], start_date, end_date, **kwargs
            )
            return prices
        raise RuntimeError(
            "Can not fetch if resource was not set during class instantiation"
        )

    @property
    def symbol(self):
        return self._symbol

    @property
    def resource(self):
        return self._resource


class Assets:
    def __init__(self, assets: List[Asset], resource: resource.Resource):
        if any(map(lambda asset: not isinstance(asset, Asset), assets)):
            raise ValueError("Expected assets to be a list of Asset")
        self._assets = assets
        self._resource = resource

    def fetch(
        self, start_date: datetime, end_date: datetime, name: str, **kwargs
    ) -> Prices:
        """
        Parameters
        ----------
        name: str
            Passed to :py:meth:`portfolio_plan.financialseries.Prices`
        """
        symbols = []
        prices = []
        for asset in self._assets:
            if asset.resource is None:
                symbols.append(asset.symbol)
            else:
                price = asset.fetch(start_date=start_date, end_date=end_date, **kwargs)
                prices.append(price)
        if symbols:
            price = self._resource.fetch(
                symbols, start_date=start_date, end_date=end_date, name=name, **kwargs
            )
            prices.append(price)
        prices_data = vstack(prices)
        return Prices(prices_data, name)

    @property
    def symbols(self) -> List[str]:
        return [asset._symbol for asset in self._assets]


class Optimizer(Enum):
    MEANVARIANCE = "MeanVariance"


class Portfolio:
    """
    A portfolio consists of assets and its associated weights
    """

    def __init__(
        self,
        assets: Assets,
        weights: Weights | None = None,
        name: str = const.DEFAULT_PORTFOLIO_NAME,
        optimizer: Optimizer = Optimizer.MEANVARIANCE,
        **optimizer_kwargs,
    ):
        """
        Parameters
        ----------
        assets: Assets
        weights: Weights | None
            Defaults to None
        name: str
            Defaults to :py:const:`portfolio_plan.const.DEFAULT_PORTFOLIO_NAME`
        optimizer: Optimizer
            Defaults to MeanVariance optimizer

        Warning
        =======
        For the moment optimizers are ignored (not implemented).

        """
        self._assets = assets
        self._weights: Weights | None = weights
        self._optimizer_kwargs = optimizer_kwargs
        self._name = name

    def fetch(
        self, start_date: datetime, end_date: datetime, name: str, **kwargs
    ) -> Prices:
        """
        Parameters
        ----------
        name: str
            Passed to :py:meth:`portfolio_plan.financialseries.Prices`
        """
        return self._assets.fetch(
            start_date=start_date, end_date=end_date, name=name, **kwargs
        )

    def _compute(self):
        """
        Compute optimization problem for optimal weights
        """
        ...

    @property
    def name(self) -> str:
        return self._name

    def weights(
        self, start_date: datetime | None = None, end_date: datetime | None = None
    ) -> Weights:
        if self._weights is not None:
            return self._weights
        else:
            self._compute()
            raise NotImplementedError()

    @property
    def symbols(self) -> List[str]:
        symbs = self._assets.symbols
        return symbs

    def plot_allocation(
        self,
        variant: Literal["main", "moon", "dawn"] = "dawn",
        start_date: datetime | None = None,
        end_date: datetime | None = None,
        base_size: int = THEME_ROSE_PINE_BASE_SIZE,
    ) -> ggplot:
        weights: Weights = self.weights(start_date=start_date, end_date=end_date)
        if weights.empty:
            symbols = self._assets.symbols
            n_symbols = len(symbols)
            weight = 1 / n_symbols
            weights = Weights(**{symbol: weight for symbol in symbols})
        return plot_allocation(weights, variant=variant, base_size=base_size)


class Plan:
    def __init__(self, portfolio: Portfolio, start_date: datetime, end_date: datetime):
        """
        Parameters
        ----------
        portfolio: Portfolio
            See :py:class:`Portfolio`
        start_date: datetime
            Start date of the plan
        end_date: datetime
            End date of the plan
        """

        self._portfolio = portfolio
        self._start_date = start_date
        self._end_date = end_date
        self._frequency: Frequency | None = None

    def __str__(self):
        name = self.__class__.__name__
        s = _create_str_stack(
            f"{name} for period {self._start_date}, {self._end_date}",
            self._portfolio.symbols,
        )
        return s

    def __repr__(self):
        return self.__str__()

    @lru_cache
    def prices(self) -> Prices:
        p = self._portfolio.fetch(
            self._start_date, self._end_date, name=self._portfolio.name
        )
        symbols = p.data.index.get_level_values("symbol").unique()
        frequencies = [pd.infer_freq(p.data.loc[symbol].index) for symbol in symbols]
        # assert frequency is the same across all symbols
        assert all(
            frequencies[i] == frequencies[i - 1] for i in range(1, len(frequencies))
        )

        freq = frequencies[0]
        self.frequency = freq  # type: ignore
        return p

    @property
    def frequency(self) -> Frequency | None:
        return self._frequency

    @frequency.setter
    def frequency(self, value: str | Frequency | None):
        if isinstance(value, Frequency):
            self._frequency = value
        elif isinstance(value, str):
            self._frequency = Frequency.from_pandas(value)
        elif value is None:
            self._frequency = None
        else:
            raise TypeError("Unsupported type")

    @lru_cache
    def returns(self, name: str) -> Returns:
        prices = self.prices()
        returns = returns_from_prices(prices.data)

        return Returns(returns, name=self._portfolio.name)

    def portfolio_returns(self) -> Returns:
        """
        Portfolio return across time
        """
        returns = self.returns(name=self._portfolio.name)
        portfolio_name = self._portfolio.name
        return returns.avg(
            self._portfolio.weights(
                start_date=self._start_date, end_date=self._end_date
            ),
            portfolio_name=portfolio_name,
        )

    @property
    def weights(self) -> Weights:
        return self._portfolio.weights(
            start_date=self._start_date, end_date=self._end_date
        )

    @property
    def start_date(self) -> datetime:
        """
        The plan start date, which is not necessarily the effective start date.
        The effective start date is the first date with available data, and it satisfies:
        effective start date >= start_date <= effective end date, where effective end date <= end date.
        """
        return self._start_date

    @property
    def end_date(self) -> datetime:
        """
        The plan end date, which is the final date of the plan period.
        """
        return self._end_date

    @property
    def portfolio(self):
        return self._portfolio

    def __rshift__(self, right):
        if not isinstance(right, self.__class__):
            type1 = type(self.right)
            type2 = self.__class__
            raise TypeError(f"unsupported operand type(s) for >>: {type1} and {type2}")

        return JoinedPlan([self, right])


def _validate_plans(plans: List[Plan]):
    if not plans:
        return
    for i in range(1, len(plans)):
        current: Plan = plans[i]
        previous: Plan = plans[i - 1]
        if current._start_date <= previous._end_date:
            raise PeriodOverlapError(
                "Plans can not overlap in time period and must be ordered by time period"
            )


class JoinedPlan:
    def __init__(self, plans=List[Plan]) -> None:
        if not plans:
            raise ValueError("At least one plan is required")
        self._plans: List[Plan] = plans
        _validate_plans(plans)

    def __str__(self):
        name = self.__class__.__name__
        return ascii_graph(name, self._plans)

    def __repr__(self):
        return self.__str__()

    def portfolio_returns(self) -> List[Returns]:
        """
        Returns each plan portfolio returns

        Returns
        -------
        A list of returns of length equal to the number of plans in the joined plan
        """
        returnss = []
        for plan in self._plans:
            returnss.append(plan.portfolio_returns())
        return returnss

    def _get_portfolio_name(self, i: int) -> str:
        return self._plans[i]._portfolio.name

    def plot_allocations(
        self,
        variant: Literal["main", "moon", "dawn"] = "dawn",
        base_size: int = THEME_ROSE_PINE_BASE_SIZE,
    ) -> ggplot:
        weights: List[Weights] = []
        for plan in self._plans:
            portfolio = plan._portfolio
            w: Weights = portfolio.weights(
                start_date=plan.start_date, end_date=plan.end_date
            )
            if w.empty:
                symbols = portfolio._assets.symbols
                n_symbols = len(symbols)
                weight = 1 / n_symbols
                w = Weights(**{symbol: weight for symbol in symbols})
            weights.append(w)
        return plot_allocations(
            weights, periods=self.periods, variant=variant, base_size=base_size
        )

    def plot_allocations2(
        self,
        variant: Literal["main", "moon", "dawn"] = "dawn",
        base_size: int = THEME_ROSE_PINE_BASE_SIZE,
        text_size: float = 0.8 * THEME_ROSE_PINE_BASE_SIZE,
        text_angle: float = 45,
    ) -> ggplot:
        weights: List[Weights] = []
        for plan in self._plans:
            portfolio = plan._portfolio
            w: Weights = portfolio.weights(
                start_date=plan.start_date, end_date=plan.end_date
            )
            if w.empty:
                symbols = portfolio._assets.symbols
                n_symbols = len(symbols)
                weight = 1 / n_symbols
                w = Weights(**{symbol: weight for symbol in symbols})
            weights.append(w)
        return plot_allocations2(
            weights,
            periods=self.periods,
            variant=variant,
            base_size=base_size,
            text_size=text_size,
            text_angle=text_angle,
        )

    def portfolio_cumulative_returns(
        self, name: str, starting_value: int = 1, keepna: bool = True
    ) -> CumulativeReturns:
        """ """
        plans_portfolio_returns = self.portfolio_returns()
        cumulatives: List[pd.DataFrame] = []
        for i in range(len(plans_portfolio_returns)):
            plan_portfolio_return = plans_portfolio_returns[i]
            portfolio_name = self._get_portfolio_name(i)

            if i == 0:
                cumulative = plan_portfolio_return.cumulative(
                    starting_value=starting_value, name=portfolio_name, keepna=keepna
                ).data
            else:
                previous_cumulative = cumulatives[i - 1]
                new_starting_value: List[float] = previous_cumulative.iloc[
                    -1, :
                ].tolist()
                cumulative = plan_portfolio_return.cumulative(
                    starting_value=new_starting_value,
                    name=plan_portfolio_return.name,
                    keepna=keepna,
                    check_base=False,
                ).data
            index: pd.MultiIndex = cumulative.index  # type: ignore
            newindex = index.set_levels([portfolio_name], level=0)
            cumulative.index = newindex
            cumulatives.append(cumulative)
        cumulatives_data = pd.concat(cumulatives, axis=0)
        return CumulativeReturns(
            cumulatives_data,
            name=name,
            check_base=False,
            periods=self.periods,
            check_frequency=False,
        )

    @property
    def periods(self) -> List[Period]:
        return [
            Period(start=plan.start_date, end=plan.end_date, frequency=plan.frequency)
            for plan in self._plans
        ]

    def information_ratio_returns(self, b_joined_plan: "JoinedPlan") -> Returns:
        """
        Compute information ratio for each of the plans

        Parameters
        ----------
        b_joined_plan: JoinedPlan
            Another joined plan

        Returns
        -------
        A single DataFrame with IR for each Period
        """
        # TODO: How to deal with the Organization
        irs = []
        for plan, bplan in zip(self._plans, b_joined_plan._plans):
            ir = InformationRatio(plan, bplan)
            irs.append(ir.period_returns(keep_index=True).data)
        return Returns(pd.concat(irs, axis=0), "IRS")

    def information_ratio_returns_cumulative(
        self, b_joined_plan: "JoinedPlan", starting_value: float = 1.0
    ) -> CumulativeReturns:
        plans = self._plans
        bplans = b_joined_plan._plans

        cumulatives: List[pd.DataFrame] = []

        for i in range(len(plans)):
            plan = plans[i]
            bplan = bplans[i]
            ir = InformationRatio(plan, bplan)
            irreturns = ir.period_returns(keep_index=True)

            if i == 0:
                cumulative = irreturns.cumulative(
                    "_", starting_value=starting_value
                ).data
            else:
                previous_cumulative = cumulatives[i - 1]
                new_starting_value = previous_cumulative.iloc[-1, :].tolist()
                cumulative = irreturns.cumulative(
                    "", starting_value=new_starting_value, check_base=False
                ).data

            cumulatives.append(cumulative)
        cumulatives_data = pd.concat(cumulatives, axis=0)
        mask = cumulatives_data.isna().all(axis=1)
        # TODO: Transform this in two straight lines instead --> Clearly show it's
        # a ratio. Not a time series.
        return CumulativeReturns(
            cumulatives_data[~mask],
            name="IR Cumulative",
            check_base=False,
            check_frequency=False,
        )



from datetime import datetime

import numpy as np
import pandas as pd
import pytest
from pandas import MultiIndex

from portfolio_plan._properties import Frequency, Weights
from portfolio_plan.data import (
    mock_cumulative_returns_path,
    mock_prices_path,
    mock_returns_path,
)
from portfolio_plan.utils import (
    Period,
    _find_gap_lines,
    cumulative_from_returns,
    plot_allocations,
    returns_from_prices,
    validate_dataframe,
)


class TestValidateDataFrame:
    def test_valid_dataframe(self):
        aapl_dates = pd.date_range("2023-01-01", periods=3, freq="D")
        msft_dates = pd.date_range("2023-01-10", periods=4, freq="D")
        index = MultiIndex.from_tuples(
            [("AAPL", d) for d in aapl_dates] + [("MSFT", d) for d in msft_dates],
            names=["symbol", "date"],
        )
        df = pd.DataFrame(
            {
                "open": [150.0, 151.0, 152.0, 210.0, 211.0, 212.0, 213.0],
                "close": [155.0, 156.0, 157.0, 215.0, 216.0, 217.0, 218.0],
                "high": [156.0, 157.0, 158.0, 216.0, 217.0, 218.0, 219.0],
                "low": [149.0, 150.0, 151.0, 209.0, 210.0, 211.0, 212.0],
            },
            index=index,
        )
        validate_dataframe(df)

    def test_not_dataframe(self):
        with pytest.raises(AssertionError, match="Data must be a pandas DataFrame"):
            validate_dataframe("not a dataframe")

    def test_empty_dataframe(self):
        df = pd.DataFrame()
        with pytest.raises(AssertionError, match="DataFrame should not be empty"):
            validate_dataframe(df)

    def test_wrong_columns(self):
        index = MultiIndex.from_tuples(
            [("AAPL", "2023-01-01")], names=["symbol", "date"]
        )
        df = pd.DataFrame({"price": [150]}, index=index)
        with pytest.raises(AssertionError, match="DataFrame must have exactly columns"):
            validate_dataframe(df)

    def test_non_multiindex(self):
        df = pd.DataFrame(
            {
                "open": [150.0],
                "close": [155.0],
                "high": [156.0],
                "low": [149.0],
            },
            index=[0],
        )
        with pytest.raises(AssertionError, match="Index must be a MultiIndex"):
            validate_dataframe(df)

    def test_wrong_index_names(self):
        index = MultiIndex.from_tuples(
            [("AAPL", "2023-01-01")], names=["ticker", "time"]
        )
        df = pd.DataFrame(
            {
                "open": [150.0],
                "close": [155.0],
                "high": [156.0],
                "low": [149.0],
            },
            index=index,
        )
        with pytest.raises(AssertionError, match="Index names must be"):
            validate_dataframe(df)

    def test_partial_nan_row(self):
        index = MultiIndex.from_tuples(
            [("AAPL", "2023-01-01")], names=["symbol", "date"]
        )
        df = pd.DataFrame(
            {
                "open": [150.0],
                "close": [None],
                "high": [156.0],
                "low": [149.0],
            },
            index=index,
        )
        with pytest.raises(
            AssertionError, match="Invalid NaN pattern found for symbol"
        ):
            validate_dataframe(df)

    def test_unsorted_dates(self):
        index = MultiIndex.from_tuples(
            [("AAPL", "2023-01-02"), ("AAPL", "2023-01-01")], names=["symbol", "date"]
        )
        df = pd.DataFrame(
            {
                "open": [152.0, 150.0],
                "close": [153.0, 155.0],
                "high": [154.0, 156.0],
                "low": [151.0, 149.0],
            },
            index=index,
        )
        with pytest.raises(AssertionError, match="Dates must be sorted for symbol"):
            validate_dataframe(df)

    def test_missing_freq(self):
        index = MultiIndex.from_tuples(
            [
                ("AAPL", pd.Timestamp("2023-01-01")),
                ("AAPL", pd.Timestamp("2023-01-03")),
                ("AAPL", pd.Timestamp("2023-01-06")),
                ("MSFT", pd.Timestamp("2023-01-02")),
                ("MSFT", pd.Timestamp("2023-01-04")),
                ("MSFT", pd.Timestamp("2023-01-08")),
            ],
            names=["symbol", "date"],
        )
        df = pd.DataFrame(
            {
                "open": [150.0, 152.0, 153.0, 210.0, 211.0, 212.0],
                "close": [155.0, 153.0, 154.0, 215.0, 214.0, 216.0],
                "high": [156.0, 154.0, 155.0, 216.0, 215.0, 217.0],
                "low": [149.0, 151.0, 152.0, 209.0, 211.0, 212.0],
            },
            index=index,
        )
        with pytest.raises(
            AssertionError,
            match="Date index must have an inferrable frequency for current symbol. Found freq=None",
        ):
            validate_dataframe(df)


def test__find_gap_lines_nan_blocks():
    """Test gap lines with NaN blocks in different positions."""
    dates = pd.date_range("2025-01-01", "2025-01-10", freq="D")

    values1 = [np.nan, np.nan, 100.0, 110.0, np.nan, 120, 130, np.nan, np.nan, 140]

    values2 = [
        100.0,
        110.0,
        np.nan,
        120.0,
        130.0,
        np.nan,
        140.0,
        np.nan,
        np.nan,
        np.nan,
    ]

    values3 = [100.0, 110.0, np.nan, np.nan, np.nan, 120.0, 130.0, 140.0, 150.0, 160.0]

    index = pd.MultiIndex.from_product(
        [["A", "B", "C"], dates], names=["symbol", "date"]
    )

    test_series = pd.Series(values1 + values2 + values3, index=index, name="close")

    result = _find_gap_lines(test_series)

    assert isinstance(result, pd.DataFrame)
    assert set(result.columns) == {"symbol", "date", "close", "group"}

    result_tuples = list(
        zip(result["symbol"], result["date"].dt.strftime("%Y-%m-%d"), result["close"])
    )

    # Expected connections
    expected_tuples = [
        ("A", "2025-01-04", 110.0),
        ("A", "2025-01-06", 120.0),
        ("A", "2025-01-07", 130.0),
        ("A", "2025-01-10", 140.0),
        ("B", "2025-01-02", 110.0),
        ("B", "2025-01-04", 120.0),
        ("B", "2025-01-05", 130.0),
        ("B", "2025-01-07", 140.0),
        ("C", "2025-01-02", 110.0),
        ("C", "2025-01-06", 120.0),
    ]

    assert result_tuples == expected_tuples


def test_empty_series():
    """Test behavior with empty series."""
    empty_series = pd.Series([], name="close")
    with pytest.raises(
        ValueError, match="Empty Series, cannot find gaps in empty data."
    ):
        _find_gap_lines(empty_series)


def test_no_gaps():
    """Test behavior with series containing no gaps."""
    # Create series with no gaps
    dates = pd.date_range("2025-01-01", "2025-01-05", freq="D")
    values = [100.0, 110.0, 120.0, 130.0, 140.0]
    index = pd.MultiIndex.from_product([["A"], dates], names=["symbol", "date"])
    series = pd.Series(values, index=index, name="close")

    result = _find_gap_lines(series)
    assert result.empty


def test_middle_nan_block():
    """Test specific case with NaN block only in the middle."""
    dates = pd.date_range("2025-01-01", "2025-01-07", freq="D")
    values = [100.0, 110.0, np.nan, np.nan, np.nan, 120.0, 130.0]

    index = pd.MultiIndex.from_product([["TSLA"], dates], names=["symbol", "date"])

    series = pd.Series(values, index=index, name="close")
    result = _find_gap_lines(series)

    result_tuples = set(
        zip(result["symbol"], result["date"].dt.strftime("%Y-%m-%d"), result["close"])
    )

    expected_tuples = {
        ("TSLA", "2025-01-02", 110.0),
        ("TSLA", "2025-01-06", 120.0),
    }

    assert result_tuples == expected_tuples

    assert len(result["group"].unique()) == 1


def test_multiple_symbols_same_dates():
    """Test handling of multiple symbols with gaps on same dates."""
    dates = pd.date_range("2025-01-01", "2025-01-05", freq="D")
    values = [100.0, np.nan, np.nan, 120.0, 130.0]

    index = pd.MultiIndex.from_product([["A", "B"], dates], names=["symbol", "date"])

    series = pd.Series(values * 2, index=index, name="close")
    result = _find_gap_lines(series)

    assert len(result["group"].unique()) == 2

    # Each symbol should have one connection
    assert len(result[result["symbol"] == "A"]) == 2
    assert len(result[result["symbol"] == "B"]) == 2


def test_plot_allocations():
    with pytest.raises(ValueError, match="Weights can not be empty"):
        plot_allocations(Weights())

    with pytest.raises(ValueError, match="Weights can not be empty"):
        plot_allocations(
            [Weights(), Weights()],
            periods=[
                Period(
                    start=datetime(2020, 1, 1),
                    end=datetime(2021, 1, 1),
                    frequency=Frequency.DAILY,
                ),
                Period(
                    start=datetime(2019, 1, 1),
                    end=datetime(2019, 2, 1),
                    frequency=Frequency.DAILY,
                ),
            ],
        )


@pytest.fixture()
def price_data() -> pd.DataFrame:
    df = pd.read_parquet(mock_prices_path())
    df = df.set_index(["symbol", "date"])
    return df


@pytest.fixture()
def returns_data() -> pd.DataFrame:
    df = pd.read_parquet(mock_returns_path())
    df = df.set_index(["symbol", "date"])
    df.columns.name = "Returns"
    return df.sort_index()


def test_returns_from_prices(price_data, returns_data):
    returns = returns_from_prices(price_data).sort_index()
    pd.testing.assert_frame_equal(
        returns,
        returns_data,
        check_exact=False,
        rtol=1e-3,
        atol=1e-8,
    )


@pytest.fixture()
def cumulative_returns_data() -> pd.DataFrame:
    df = pd.read_parquet(mock_cumulative_returns_path())
    df = df.set_index(["symbol", "date"])
    df.columns.name = "CumulativeReturns"
    return df


class TestCumulativeFromReturns:
    """mock data"""

    data: pd.DataFrame

    @classmethod
    def setup_class(cls):
        data = pd.read_parquet(mock_returns_path())
        data = data.set_index(["symbol", "date"])
        cls.data = data

    def test_full_returns(self, cumulative_returns_data):
        result = cumulative_from_returns(self.data, keepna=True)
        pd.testing.assert_frame_equal(
            result, cumulative_returns_data, check_exact=False, rtol=1e-3, atol=1e-3
        )

    def test_full_returns_nokeepna(self, cumulative_returns_data):
        result = cumulative_from_returns(self.data, keepna=False)
        pd.testing.assert_frame_equal(result, cumulative_returns_data.ffill())

    def test_singleton_returns(self):
        self.data

    def test_cumulative_from_returns_with_nan_handling_and_starting_value(self):
        portfolio_returns = pd.DataFrame(
            {"close": [np.nan, 0.1, 0.05]},
            index=pd.MultiIndex.from_tuples(
                [
                    ("A", pd.Timestamp("2020-01-01")),
                    ("A", pd.Timestamp("2020-01-02")),
                    ("A", pd.Timestamp("2020-01-03")),
                ],
                names=["symbol", "date"],
            ),
        )

        result_default = cumulative_from_returns(portfolio_returns)

        expected_default = pd.DataFrame(
            {"close": [1.0, 1.0 * (1.0 + 0.1), 1.0 * (1.0 + 0.1) * (1.0 + 0.05)]},
            index=pd.MultiIndex.from_tuples(
                [
                    ("A", pd.Timestamp("2020-01-01")),
                    ("A", pd.Timestamp("2020-01-02")),
                    ("A", pd.Timestamp("2020-01-03")),
                ],
                names=["symbol", "date"],
            ),
        )
        expected_default.columns.name = "CumulativeReturns"

        pd.testing.assert_frame_equal(result_default, expected_default)

        result_zero = cumulative_from_returns(portfolio_returns, starting_value=0.0)

        expected_zero = pd.DataFrame(
            {
                "close": [
                    0.0,
                    ((1.0) * (1.0 + 0.1)) - 1,
                    ((1.0) * (1.0 + 0.1) * (1.0 + 0.05)) - 1.0,
                ]
            },
            index=pd.MultiIndex.from_tuples(
                [
                    ("A", pd.Timestamp("2020-01-01")),
                    ("A", pd.Timestamp("2020-01-02")),
                    ("A", pd.Timestamp("2020-01-03")),
                ],
                names=["symbol", "date"],
            ),
        )
        expected_zero.columns.name = "CumulativeReturns"

        pd.testing.assert_frame_equal(result_zero, expected_zero)

    def test_cumulative_from_returns_with_multiple_symbols(self):
        """Test that cumulative_from_returns works correctly with multiple symbols."""

        portfolio_returns = pd.DataFrame(
            {"close": [np.nan, 0.1, 0.05, np.nan, 0.2, -0.05]},
            index=pd.MultiIndex.from_tuples(
                [
                    ("A", pd.Timestamp("2020-01-01")),
                    ("A", pd.Timestamp("2020-01-02")),
                    ("A", pd.Timestamp("2020-01-03")),
                    ("B", pd.Timestamp("2020-01-01")),
                    ("B", pd.Timestamp("2020-01-02")),
                    ("B", pd.Timestamp("2020-01-03")),
                ],
                names=["symbol", "date"],
            ),
        )

        result = cumulative_from_returns(portfolio_returns)

        expected = pd.DataFrame(
            {
                "close": [
                    1.0,
                    1.0 * 1.1,
                    1.0 * 1.1 * 1.05,  # A values
                    1.0,
                    1.0 * 1.2,
                    1.0 * 1.2 * 0.95,  # B values
                ]
            },
            index=pd.MultiIndex.from_tuples(
                [
                    ("A", pd.Timestamp("2020-01-01")),
                    ("A", pd.Timestamp("2020-01-02")),
                    ("A", pd.Timestamp("2020-01-03")),
                    ("B", pd.Timestamp("2020-01-01")),
                    ("B", pd.Timestamp("2020-01-02")),
                    ("B", pd.Timestamp("2020-01-03")),
                ],
                names=["symbol", "date"],
            ),
        )
        expected.columns.name = "CumulativeReturns"

        pd.testing.assert_frame_equal(result, expected)

    def test_cumulative_from_returns_without_keepna(self):
        """Test that cumulative_from_returns works correctly when keepna=False."""

        portfolio_returns = pd.DataFrame(
            {"close": [np.nan, 0.1, 0.05]},
            index=pd.MultiIndex.from_tuples(
                [
                    ("A", pd.Timestamp("2020-01-01")),
                    ("A", pd.Timestamp("2020-01-02")),
                    ("A", pd.Timestamp("2020-01-03")),
                ],
                names=["symbol", "date"],
            ),
        )

        result = cumulative_from_returns(portfolio_returns, keepna=False)
        print(result)

        expected = pd.DataFrame(
            {"close": [1.0, (1.0 + 0.1), (1.0 + 0.1) * (1.0 + 0.05)]},
            index=pd.MultiIndex.from_tuples(
                [
                    ("A", pd.Timestamp("2020-01-01")),
                    ("A", pd.Timestamp("2020-01-02")),
                    ("A", pd.Timestamp("2020-01-03")),
                ],
                names=["symbol", "date"],
            ),
        )
        expected.columns.name = "CumulativeReturns"

        # Check that result matches expected
        pd.testing.assert_frame_equal(result, expected)



import numpy as np
import pandas as pd
import pytest

from portfolio_plan._properties import Frequency
from portfolio_plan.financialseries import Prices, Returns, ValidationError


class TestPrices:
    @classmethod
    def setup_class(cls):
        dates = pd.date_range(start="2024-01-01", periods=7, freq="D")
        symbols = ["A"]
        idx = pd.MultiIndex.from_product([symbols, dates], names=["symbol", "date"])

        cls.data = pd.DataFrame(
            {
                "open": [100, 102, 101, 105, 104, 106, 107],
                "close": [102, 101, 105, 104, 106, 107, 109],
                "high": [103, 103, 106, 107, 107, 108, 110],
                "low": [99, 100, 100, 103, 103, 105, 106],
            },
            index=idx,
        )

        cls.instance = Prices(data=cls.data, name="_")

    def test_resample_daily(self):
        frequency = Frequency.DAILY
        resampled = self.instance.resample_period(frequency)

        pd.testing.assert_frame_equal(resampled._data, self.data)

    def test_resample_weekly(self):
        frequency = Frequency.WEEKLY
        resampled = self.instance.resample_period(frequency)

        expected_idx = pd.MultiIndex.from_tuples(
            [
                ("A", pd.Timestamp("2024-01-07")),
            ],
            names=["symbol", "date"],
        )

        expected = pd.DataFrame(
            {
                "open": [100],
                "close": [109],
                "high": [110],
                "low": [99],
            },
            index=expected_idx,
        )

        pd.testing.assert_frame_equal(resampled._data, expected)

    def test_resample_minute_raises(self):
        frequency = Frequency.MINUTE
        with pytest.raises(ValueError, match="Sampling from D to 1T is not allowed."):
            self.instance.resample_period(frequency)


class TestReturns:
    @classmethod
    def setup_class(cls):
        dates = pd.date_range(start="2024-01-01", periods=4, freq="D")
        symbols = ["A", "B"]
        idx = pd.MultiIndex.from_product([symbols, dates], names=["symbol", "date"])

        cls.data = pd.DataFrame(
            {
                "open": [np.nan, 0.2, 0.3, 0.2, np.nan, 0.1, 0.2, 0.1],
                "close": [np.nan, 0.2, 0.1, 0.3, np.nan, 0.5, 0.1, 0.2],
                "high": [np.nan, 0.1, 0.1, 0.3, np.nan, 0.5, 0.1, 0.2],
                "low": [np.nan, 0.2, 0.1, 0.3, np.nan, 0.5, 0.1, 0.4],
            },
            index=idx,
        )

        cls.instance = Returns(data=cls.data, name="_")

    @staticmethod
    def test_instantiation():
        dates = pd.date_range(start="2024-01-01", periods=3, freq="D")
        symbols = ["A", "B"]
        idx = pd.MultiIndex.from_product([symbols, dates], names=["symbol", "date"])

        data = pd.DataFrame(
            {
                "open": [0.2, 0.3, 0.2, 0.1, 0.2, 0.1],
                "close": [0.2, 0.1, 0.3, 0.5, 0.1, 0.2],
                "high": [0.1, 0.1, 0.3, 0.5, 0.1, 0.2],
                "low": [0.2, 0.1, 0.3, 0.5, 0.1, 0.4],
            },
            index=idx,
        )
        with pytest.raises(
            ValidationError,
        ):
            Returns(data=data, name="_")

    def test_cumulative(self):
        result = self.instance.cumulative("_").data.sort_index(level=0)

        expected_index = pd.MultiIndex.from_product(
            [["A", "B"], pd.date_range("2024-01-01", periods=4)],
            names=["symbol", "date"],
        )

        expected = pd.DataFrame(
            {
                "open": [
                    1,
                    1.2,
                    1.2 * 1.3,
                    1.2 * 1.3 * 1.2,
                    1,
                    1.1,
                    1.1 * 1.2,
                    1.1 * 1.2 * 1.1,
                ],
                "close": [
                    1,
                    1.2,
                    1.2 * 1.1,
                    1.2 * 1.1 * 1.3,
                    1,
                    1.5,
                    1.5 * 1.1,
                    1.5 * 1.1 * 1.2,
                ],
                "high": [
                    1,
                    1.1,
                    1.1 * 1.1,
                    1.1 * 1.1 * 1.3,
                    1,
                    1.5,
                    1.5 * 1.1,
                    1.5 * 1.1 * 1.2,
                ],
                "low": [
                    1,
                    1.2,
                    1.2 * 1.1,
                    1.2 * 1.1 * 1.3,
                    1,
                    1.5,
                    1.5 * 1.1,
                    1.5 * 1.1 * 1.4,
                ],
            },
            index=expected_index,
        )
        expected.columns.name = "CumulativeReturns"

        pd.testing.assert_frame_equal(result, expected)



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
from empyrical import excess_sharpe

if TYPE_CHECKING:
    from portfolio_plan.core import Plan

from portfolio_plan.financialseries import Returns
from portfolio_plan.utils import add_singleton_level


class DifferentIndexError(ValueError):
    pass


def _validate_information_ratio_input(
    returns: pd.DataFrame, benchmark_returns: pd.DataFrame
):
    if isinstance(returns.index, pd.MultiIndex):
        raise ValueError("Expected single index")
    if isinstance(benchmark_returns.index, pd.MultiIndex):
        raise ValueError("Expected single index")
    if not benchmark_returns.index.equals(returns.index):
        raise DifferentIndexError(
            "returns and benchmark_returns do not share same index values"
        )
    assert returns.index.is_monotonic_increasing


def information_ratio(
    returns: pd.DataFrame,
    benchmark_returns: pd.DataFrame,
    name: str = "",
    keep_index: bool = False,
) -> pd.DataFrame:
    """
    Compute the non-annualized information ratio for all OHLC columns


    Parameters
    ----------
    returns: pd.DataFrame
        OHLC returns of one asset
    benchmark_returns: pd.DataFrame
        OHLC returns of one benchmark asset

    Returns
    -------
    A single index DataFrame with the IR for each dimension of OHLC
    """
    # NOTE: empyrical.excess_sharpe is an information ratio with no annualization
    try:
        _validate_information_ratio_input(returns, benchmark_returns)
    except DifferentIndexError as e:
        if returns.index.shape[0] > benchmark_returns.index.shape[0]:
            # TODO: check what are the values that are present in the other
            if not set(benchmark_returns.index).issubset(returns.index):
                raise e
            benchmark_returns = benchmark_returns.reindex(returns.index)
        elif returns.index.shape[0] < benchmark_returns.index.shape[0]:
            if not set(returns.index).issubset(returns.index):
                raise e
            returns = returns.reindex(benchmark_returns.index)
        else:
            raise e

    if keep_index:
        nrow = max(1, returns.shape[0])
        index = returns.index
    else:
        nrow = 1
        max_date = returns.index.get_level_values("date").max()
        index = pd.Index([max_date])

    out = np.empty((nrow, 4), dtype=np.float64)
    out[:] = np.nan
    columns = ["open", "low", "high", "close"]
    for i, dim in enumerate(columns):
        r = returns[dim].to_numpy()
        b = benchmark_returns[dim].to_numpy()
        excess_sharpe(r, b, out=out[-1:, i])

    return pd.DataFrame(out, columns=columns, index=index)


class InformationRatio:
    def __init__(self, reference: Plan, benchmark: Plan):
        self._reference = reference
        self._benchmark = benchmark

    def period_returns(self, keep_index: bool = False) -> Returns:
        """
        Parameters
        ---------
        keep_index: bool
            See :py:func:`information_ratio`
        Returns
        -------
            Returns, a singleton return
        """
        portfolio_returns = self._reference.portfolio_returns()
        benchmark_returns = self._benchmark.portfolio_returns()
        assert portfolio_returns.is_unique_symbol
        assert benchmark_returns.is_unique_symbol
        pdata = portfolio_returns.data.droplevel(level="symbol")
        bdata = benchmark_returns.data.droplevel(level="symbol")

        ir = information_ratio(pdata, bdata, keep_index=keep_index)
        portfolio_returns.data.index.get_level_values("symbol")
        name = f"IR: {portfolio_returns.name}-{benchmark_returns.name}"
        add_singleton_level(ir, name)
        ir.name = "IR"
        return Returns(ir, "IR")



from collections import UserDict
from enum import Enum

from portfolio_plan.errors import ValidationError


class Weights(UserDict):
    def __init__(self, **kwargs):
        """
        A user-defined dictionary containing symbols and associated weights.

        Example
        -------
        >>> weights = Weights(A=0.2, B=0.8)
        >>> weights

        >>> a_weight = weights["A"]

        # For more complicated symbols
        >>> weights = Weights(**{"A.S": 1})

        # No parameters passed to a weights object will be considered,
        # by convention as an equal weight
        >>> weights = Weights()
        """
        data = dict(kwargs)
        self._validate(data)
        super().__init__(data)

    @staticmethod
    def _validate(d: dict):
        if not d:
            return

        if not all(isinstance(v, (int, float)) for v in d.values()):
            raise ValidationError("All weights must be numeric (int or float).")

        total = sum(d.values())
        if not abs(total - 1.0) < 1e-8:
            raise ValidationError(f"Weights must sum to 1. Provided sum: {total}")

    def __repr__(self):
        items = ", ".join(f"{k}: {v:.4f}" for k, v in self.items())
        return f"Weights({{{items}}})"

    @property
    def empty(self) -> bool:
        return len(self) == 0


class Frequency(Enum):
    MINUTE = "1m"
    DAILY = "1d"
    BDAILY = "1bd"
    WEEKLY = "1wk"
    MONTHLY = "1mo"

    def to_pandas(self):
        # TODO: support pandas>=2.0.0
        match self:
            case Frequency.MINUTE:
                return "1T"
            case Frequency.DAILY:
                return "1D"  # day
            case Frequency.BDAILY:
                return "1B"
            case Frequency.WEEKLY:
                return "1W"  # week
            case Frequency.MONTHLY:
                return "1M"  # TODO: check this
            case _:
                raise ValueError(f"Unknown frequency: {self}")

    @staticmethod
    def from_pandas(value: str):
        match value:
            case "T":
                return Frequency.MINUTE
            case "D":
                return Frequency.DAILY
            case "B":
                return Frequency.BDAILY
            case "W":
                return Frequency.WEEKLY
            case "M":
                return Frequency.MONTHLY  # TODO: Check this
            case _:
                raise ValueError(f"Unknown pandas frequency: {value}")

    def to_yf(self):
        match self:
            case (
                Frequency.MINUTE
                | Frequency.DAILY
                | Frequency.WEEKLY
                | Frequency.MONTHLY
            ):
                return self.value

            case Frequency.BDAILY:
                return Frequency.DAILY.value
            case _:
                raise ValueError(f"Unknown frequency {self}")



from io import BytesIO
from typing import Dict, List, Tuple

import numpy as np
import pandas as pd
import pytest

from portfolio_plan import const
from portfolio_plan._properties import Frequency, Weights
from portfolio_plan.core import Asset, Assets, Plan, Portfolio
from portfolio_plan.errors import PeriodOverlapError
from portfolio_plan.financialseries import ValidationError
from portfolio_plan.resource import Memory as ArrowMemory
from portfolio_plan.utils import (
    add_singleton_level,
    cumulative_from_returns,
    returns_from_prices,
)


class TestWeight:
    @staticmethod
    def test_valid_weights():
        w = Weights(AAPL=0.6, MSFT=0.4)
        assert w["AAPL"] == 0.6
        assert w["MSFT"] == 0.4

    @staticmethod
    def test_valid_weights_complex_key():
        w = Weights(**{"NESN.SW": 1.0})
        assert w["NESN.SW"] == 1.0

    @staticmethod
    def test_sum_not_one():
        with pytest.raises(ValidationError):
            Weights(AAPL=0.5, MSFT=0.3)

    @staticmethod
    def test_empty_weights():
        # Does not raise error
        Weights()

    @staticmethod
    def test_non_numeric_weights():
        with pytest.raises(ValidationError):
            Weights(AAPL="high", MSFT=0.5)

    @staticmethod
    def test_floating_point_tolerance():
        w = Weights(AAPL=0.3333333, MSFT=0.6666667)
        assert abs(sum(w.values()) - 1.0) < 1e-8

    @staticmethod
    def test_repr_output():
        w = Weights(AAPL=0.5, GOOG=0.5)
        rep = repr(w)
        assert "AAPL" in rep
        assert "GOOG" in rep


def setup_data(
    data: Dict[str, list[float | int]], symbols: List[str], start: str, freq: str = "D"
) -> pd.DataFrame:
    nobs = len(data["open"])
    nentity = len(symbols)

    assert nobs % nentity == 0
    nperiod = nobs // nentity
    date = pd.date_range(start=start, periods=nperiod, freq=freq)
    index = pd.MultiIndex.from_product([symbols, date], names=["symbol", "date"])

    df = pd.DataFrame(data, index=index)
    return df


def setup_plan(data: pd.DataFrame) -> Plan:
    symbols: List[str] = data.index.get_level_values(level="symbol").unique().tolist()
    dates = data.index.get_level_values(level="date")
    stream = BytesIO()
    data.to_parquet(stream, engine="pyarrow")
    plan = Plan(
        portfolio=Portfolio(
            assets=Assets(
                [Asset(symbol, resource=None) for symbol in symbols],
                resource=ArrowMemory(
                    data=stream,
                    frequency=Frequency.DAILY,
                ),
            ),
            weights=Weights(),
        ),
        start_date=dates.min(),
        end_date=dates.max(),
    )
    return plan


@pytest.fixture()
def input_mixed() -> Tuple[pd.DataFrame, pd.DataFrame]:
    return (
        setup_data(
            {
                "open": [
                    100.0,
                    102.0,
                    101.0,
                    105.0,
                    104.0,
                    106.0,
                    107.0,
                    110.0,
                    108.0,
                    109.0,
                    111.0,
                    112.0,
                    113.0,
                    115.0,
                ],
                "close": [
                    102.0,
                    101.0,
                    105.0,
                    104.0,
                    106.0,
                    107.0,
                    109.0,
                    112.0,
                    110.0,
                    111.0,
                    114.0,
                    115.0,
                    116.0,
                    118.0,
                ],
                "high": [
                    103.0,
                    103.0,
                    106.0,
                    107.0,
                    107.0,
                    108.0,
                    110.0,
                    113.0,
                    111.0,
                    112.0,
                    115.0,
                    116.0,
                    117.0,
                    119.0,
                ],
                "low": [
                    99.0,
                    100.0,
                    100.0,
                    103.0,
                    103.0,
                    105.0,
                    106.0,
                    109.0,
                    107.0,
                    108.0,
                    110.0,
                    111.0,
                    112.0,
                    114.0,
                ],
            },
            symbols=["A", "B"],
            start="2024-01-01",
        ),
        setup_data(
            {
                "open": [200.0, 202.0, 204.0, 206.0, 208.0, 210.0],
                "close": [202.0, 204.0, 206.0, 208.0, 210.0, 212.0],
                "high": [203.0, 205.0, 207.0, 209.0, 211.0, 213.0],
                "low": [199.0, 201.0, 203.0, 205.0, 207.0, 209.0],
            },
            symbols=["B", "C"],
            start="2024-01-08",
        ),
    )


def _expected_returns(input: Tuple[pd.DataFrame, pd.DataFrame]) -> List[pd.DataFrame]:
    price1, price2 = input
    returns1 = returns_from_prices(price1)
    returns2 = returns_from_prices(price2)

    # Equal weight portfolio
    average_returns1 = returns1.groupby(level="date").mean()
    average_returns2 = returns2.groupby(level="date").mean()
    add_singleton_level(average_returns1, const.DEFAULT_PORTFOLIO_NAME)
    add_singleton_level(average_returns2, const.DEFAULT_PORTFOLIO_NAME)
    return [average_returns1, average_returns2]


def _expected_cumulative_returns(average_returns1, average_returns2) -> pd.DataFrame:
    cumulative_returns1 = cumulative_from_returns(average_returns1, 1, True)
    cumulative_returns2 = []
    for column in ("close", "open", "high", "low"):
        last = cumulative_returns1.loc[const.DEFAULT_PORTFOLIO_NAME][column].iloc[-1]
        assert not pd.isna(last)
        _ = cumulative_from_returns(average_returns2[[column]], last, True)
        cumulative_returns2.append(_)
    cumulative_returns2 = pd.concat(cumulative_returns2, axis=1)  # type: ignore
    exp = pd.concat([cumulative_returns1, cumulative_returns2], axis=0)  # type: ignore

    return exp


@pytest.fixture()
def expected_returns_mixed(
    input_mixed: Tuple[pd.DataFrame, pd.DataFrame],
) -> list[pd.DataFrame]:
    return _expected_returns(input_mixed)


@pytest.fixture()
def expected_cumulative_returns_mixed(
    input_mixed: Tuple[pd.DataFrame, pd.DataFrame],
) -> pd.DataFrame:
    avg1, avg2 = _expected_returns(input_mixed)
    return _expected_cumulative_returns(avg1, avg2)


@pytest.fixture
def input_distinct() -> Tuple[pd.DataFrame, pd.DataFrame]:
    return (
        setup_data(
            {
                "open": [100.0, 101.0, 102.0],
                "close": [101.0, 103.0, 104.0],
                "high": [102.0, 104.0, 105.0],
                "low": [99.0, 100.0, 101.0],
            },
            symbols=["A"],
            start="2024-01-01",
        ),
        setup_data(
            {
                "open": [200.0, 205.0, 208.0],
                "close": [205.0, 210.0, 212.0],
                "high": [206.0, 211.0, 213.0],
                "low": [198.0, 204.0, 206.0],
            },
            symbols=["B"],
            start="2024-01-04",
        ),
    )


@pytest.fixture
def expected_returns_distinct(input_distinct) -> list[pd.DataFrame]:
    return _expected_returns(input_distinct)


@pytest.fixture
def expected_cumulative_returns_distinct(input_distinct) -> pd.DataFrame:
    avg1, avg2 = _expected_returns(input_distinct)
    return _expected_cumulative_returns(avg1, avg2)


@pytest.fixture()
def input_overlapping_period() -> Tuple[pd.DataFrame, pd.DataFrame]:
    return (
        setup_data(
            {
                "open": [100.0, 102.0],
                "close": [102.0, 104.0],
                "high": [103.0, 105.0],
                "low": [99.0, 101.0],
            },
            symbols=["A"],
            start="2024-01-01",
        ),
        setup_data(
            {
                "open": [105.0, 106.0],
                "close": [106.0, 108.0],
                "high": [107.0, 109.0],
                "low": [104.0, 105.0],
            },
            symbols=["A"],
            start="2024-01-02",
        ),
    )


class TestPlan:
    def test_join_returns_mixed(
        self, input_mixed: Tuple[pd.DataFrame, pd.DataFrame], expected_returns_mixed
    ):
        price1, price2 = input_mixed

        plan1 = setup_plan(price1)
        plan2 = setup_plan(price2)

        joined = plan1 >> plan2
        result = joined.portfolio_returns()
        assert len(result) == len(expected_returns_mixed)
        for a, b in zip(result, expected_returns_mixed):
            pd.testing.assert_frame_equal(a.data, b)

    def test_join_cumulative_returns_mixed(
        self,
        input_mixed: Tuple[pd.DataFrame, pd.DataFrame],
        expected_cumulative_returns_mixed,
    ):
        price1, price2 = input_mixed

        plan1 = setup_plan(price1)
        plan2 = setup_plan(price2)

        joined = plan1 >> plan2
        result = joined.portfolio_cumulative_returns("_")
        pd.testing.assert_frame_equal(result.data, expected_cumulative_returns_mixed)

    def test_join_returns_distinct(
        self,
        input_distinct: Tuple[pd.DataFrame, pd.DataFrame],
        expected_returns_distinct: list[pd.DataFrame],
    ):
        price1, price2 = input_distinct

        plan1 = setup_plan(price1)
        plan2 = setup_plan(price2)

        joined = plan1 >> plan2

        result = joined.portfolio_returns()
        assert len(result) == len(expected_returns_distinct)
        for a, b in zip(result, expected_returns_distinct):
            pd.testing.assert_frame_equal(a.data, b)

    def test_join_cumulative_returns_distinct(
        self,
        input_distinct: Tuple[pd.DataFrame, pd.DataFrame],
        expected_cumulative_returns_distinct,
    ):
        price1, price2 = input_distinct

        plan1 = setup_plan(price1)
        plan2 = setup_plan(price2)

        joined = plan1 >> plan2

        result = joined.portfolio_cumulative_returns("_")
        pd.testing.assert_frame_equal(result.data, expected_cumulative_returns_distinct)

    def test_overlapping_period(
        self, input_overlapping_period: Tuple[pd.DataFrame, pd.DataFrame]
    ):
        price1, price2 = input_overlapping_period

        plan1 = setup_plan(price1)
        plan2 = setup_plan(price2)

        with pytest.raises(PeriodOverlapError):
            _ = plan1 >> plan2

    def test_portfolio_returns(self):
        """
        Portfolio return across time
        """
        values = [
            100.0,  # A, 2020-01-01
            110.0,  # A, 2020-01-02
            115.0,  # A, 2020-01-03
            120.0,  # A, 2020-01-04
            200.0,  # B, 2020-01-01
            210.0,  # B, 2020-01-02
            220.0,  # B, 2020-01-03
            225.0,  # B, 2020-01-04
        ]

        index_tuples = [
            ("A", pd.Timestamp("2020-01-01")),
            ("A", pd.Timestamp("2020-01-02")),
            ("A", pd.Timestamp("2020-01-03")),
            ("A", pd.Timestamp("2020-01-04")),
            ("B", pd.Timestamp("2020-01-01")),
            ("B", pd.Timestamp("2020-01-02")),
            ("B", pd.Timestamp("2020-01-03")),
            ("B", pd.Timestamp("2020-01-04")),
        ]

        data = pd.DataFrame(
            {
                "open": values,
                "close": values,
                "high": values,
                "low": values,
            },
            index=pd.MultiIndex.from_tuples(
                index_tuples,
                names=["symbol", "date"],
            ),
        )

        plan = setup_plan(data)
        result = plan.portfolio_returns()

        expected_values = [
            np.nan,  # 2020-01-01
            (((110.0 - 100.0) / 100.0) + ((210.0 - 200.0) / 200.0)) / 2,  # 2020-01-02
            (((115.0 - 110.0) / 110.0) + ((220.0 - 210.0) / 210.0)) / 2,  # 2020-01-03
            (((120.0 - 115.0) / 115.0) + ((225.0 - 220.0) / 220.0)) / 2,  # 2020-01-04
        ]

        expected = pd.DataFrame(
            {
                "open": expected_values,
                "close": expected_values,
                "high": expected_values,
                "low": expected_values,
            },
            index=pd.to_datetime(
                ["2020-01-01", "2020-01-02", "2020-01-03", "2020-01-04"]
            ),
        )
        add_singleton_level(expected, "Portfolio")
        expected.columns.name = "Returns"

        pd.testing.assert_frame_equal(result.data, expected)



from typing import List, Literal

import numpy as np
import pandas as pd
from plotnine.ggplot import ggplot

from portfolio_plan import const
from portfolio_plan._properties import Frequency, Weights
from portfolio_plan.abc import FinancialSeries
from portfolio_plan.errors import ValidationError
from portfolio_plan.utils import (
    Period,
    _geom_lines,
    _geom_text,
    add_singleton_level,
    cumulative_from_returns,
    plot_cumulative_returns,
    plot_prices,
    plot_returns,
    validate_dataframe,
)


class Prices(FinancialSeries):
    def __init__(self, data: pd.DataFrame, name: str):
        """
        Parameters
        ----------
        data: pd.DataFrame
            The data including financial prices.
        name: str
            The name of this set of financial prices.

        Note
        ----
        It is recommended to derive `Prices` instance from a resource rather
        than from direct instantiation.
        A resource, can be a file, an in memory parquet buffer, a network resource.
        Refer to :ref:`resource API`
        """

        super().__init__(data, name)

    def validate_data(self, data: pd.DataFrame):
        validate_dataframe(data)

    def resample_period(self, frequency: Frequency):
        ohlc_dict = {
            "open": "first",
            "close": "last",
            "high": "max",
            "low": "min",
        }
        usymbols = self.data.index.get_level_values("symbol").unique()
        frequencies = [
            pd.infer_freq(self.data.loc[symbol].index) for symbol in usymbols
        ]
        assert all(
            frequencies[i] == frequencies[i - 1] for i in range(1, len(frequencies))
        )
        current_freq = frequencies[0]
        if current_freq is None:
            raise ValueError("Cannot infer current frequency of the data.")

        def freq_to_timedelta(freq):
            match freq:
                case "1B":
                    approx_timedelta = pd.Timedelta(days=1.4) * 5
                    return approx_timedelta
                case "1D" | "1W" | "1M" | "1T":
                    pass
                    return pd.to_timedelta(freq)
                case _:
                    raise ValueError(f"Upsampling of {current_freq} is not supported.")

        current_td = freq_to_timedelta("1" + current_freq)

        target_td = freq_to_timedelta(frequency.to_pandas())
        target_freq = frequency.to_pandas()
        if target_td < current_td:
            raise ValueError(
                f"""Sampling from {current_freq} to {target_freq} is not allowed.
                Down sampling is not supported"""
            )

        resampled = (
            self._data.groupby(level="symbol")
            .resample(target_freq, level="date")
            .agg(ohlc_dict)  # type: ignore
        )
        resampled.index = pd.MultiIndex.from_tuples(
            resampled.index.to_flat_index(), names=["symbol", "date"]
        )
        return self.__class__(data=resampled, name=self.name)

    def plot_low(self, scale_x: bool = True) -> ggplot:
        """
        Plot all prices low price

        Example
        -------

        >>> from portfolio_plan.financialseries import Prices, Frequency
        >>> from portfolio_plan.resource import File
        >>> from portfolio_plan.data import example_prices_path


        >>> resource = File(path=example_prices_path(), frequency=Frequency.BDAILY)
        >>> prices = resource.fetch(
        ...     ["A", "B"],
        ...     start_date="2025-01-01",
        ... end_date="2025-01-31"
        )
        >>> price.plot_low()
        """
        return plot_prices(self._data["low"], scale_x)

    def plot_high(self, scale_x: bool = True) -> ggplot:
        """
        Plot all prices high price
        """
        return plot_prices(self._data["high"], scale_x)

    def plot_open(self, scale_x: bool = True) -> ggplot:
        """
        Plot all prices open price
        """

        return plot_prices(self._data["open"], scale_x)

    def plot_close(self, scale_x: bool = True) -> ggplot:
        """
        Plot all price close price
        """
        return plot_prices(self._data["close"], scale_x)

    def plot(self, scale_x: bool = True):
        """
        Plot all prices candle chart
        """
        raise NotImplementedError()


class CumulativeReturns(FinancialSeries):
    def __init__(
        self,
        data: pd.DataFrame,
        name: str,
        check_base: bool = True,
        periods: List[Period] | None = None,
        check_frequency: bool = True,
    ):
        """
        Parameters
        ----------
        data: pd.DataFrame
            A dataframe containing cumulative OHLC returns
        name: str
            Name of the cumulative returns
        check_base: bool
            Whether to enforce first value to be equal across open, close, high, low.
            Defaults to True
        periods: List[Period] | None
            See :py:meth:`portfolio_plan.utils.Period`
        """
        self._check_frequency = check_frequency
        super().__init__(data, name)
        subset = data.groupby("symbol").nth(0)
        if check_base:
            if not np.all(subset.values == subset.values[0, 0]):
                raise ValidationError(
                    "Cumulative Returns should always start with same base value across symbols."
                )
        self._periods = periods

    def validate_data(self, data: pd.DataFrame):
        validate_dataframe(data, check_frequency=self._check_frequency)

    def plot_low(self, scale_x: bool = True) -> ggplot:
        """
        Plot all cumulative returns high price
        """
        return plot_cumulative_returns(self._data["low"], scale_x, periods=self.periods)

    def plot_high(self, scale_x: bool = True) -> ggplot:
        """
        Plot all cumulative returns high price
        """
        return plot_cumulative_returns(
            self._data["high"], scale_x, periods=self.periods
        )

    def plot_open(self, scale_x: bool = True) -> ggplot:
        """
        Plot all cumulative returns open price
        """

        return plot_cumulative_returns(
            self._data["open"], scale_x, periods=self.periods
        )

    def plot_close(self, scale_x: bool = True) -> ggplot:
        """
        Plot all cumulative returns close price
        """
        return plot_cumulative_returns(
            self._data["close"], scale_x, periods=self.periods
        )

    def plot(self, scale_x: bool = True):
        """
        Plot all cumulative returns candle chart
        """
        raise NotImplementedError()

    @property
    def periods(self) -> List[Period] | None:
        return self._periods


class Returns(FinancialSeries):
    def __init__(self, data: pd.DataFrame, name: str):
        """
        Parameters
        ----------
        data: pd.DataFrame
            A dataframe containing the OHLC returns
        """
        super().__init__(data, name)
        subset = data.groupby(level="symbol").nth(0)
        if not self.is_singleton:
            if not subset.isna().all(axis=1).all():  # type: ignore
                raise ValidationError(
                    "Each symbols returns are expected to have first row made of NaNs"
                )

    def validate_data(self, data: pd.DataFrame):
        validate_dataframe(data)

    def avg(
        self, weights: Weights, portfolio_name: str = const.DEFAULT_PORTFOLIO_NAME
    ) -> "Returns":
        """
        Compute the average return by date
        """
        weighted = []
        if not len(weights):
            weighted_avg = self._data.groupby(level="date").mean()
            add_singleton_level(weighted_avg, portfolio_name)
            return self.__class__(weighted_avg, name=self.name)

        for symbol, weight in weights.items():
            w = self._data.loc[symbol, :] * weight
            w.name = symbol
            weighted.append(w)

        weighted_data = pd.concat(
            weighted,
            axis=0,
            keys=[df.name for df in weighted],
            names=["symbol", "date"],
        )
        namask = weighted_data.xs(symbol, level="symbol").isna().all(axis=1)
        weighted_avg = weighted_data.groupby(level="date").sum()
        weighted_avg[namask] = np.nan

        add_singleton_level(weighted_avg, portfolio_name)
        return self.__class__(weighted_avg, name=self.name)

    def cumulative(
        self,
        name: str,
        starting_value: float | List[float] = 1.0,
        keepna: bool = True,
        check_base: bool = True,
        periods: List[Period] | None = None,
    ) -> CumulativeReturns:
        """
        Parameters
        ----------
        name: str
            See :py:meth:`portfolio_plan.financialseries.CumulativeReturns`
        starting_value: int | List[int]
            See :py:func:`empyrical.cum_returns`
            If starting_value is a list, it must be of length 4.
            Each element is mapped (in order) to:
            - open
            - close
            - high
            - low

        Warning
        -------
        I recommend setting keepna to True. When False, inconsistencies between
        the frequencies of the scale of the plotted dates and the actual frequency
        of the data could lead to misinterpretation.
        """
        if not isinstance(starting_value, list):
            newreturns: pd.DataFrame = cumulative_from_returns(
                self._data, starting_value, keepna
            )
        else:
            if len(starting_value) != 4:
                raise ValueError("Expected starting_value to be of length 4")
            columns = ("open", "close", "high", "low")
            newreturnss: List[pd.DataFrame] = []
            for i, value in enumerate(starting_value):
                column = columns[i]
                newreturnss.append(
                    cumulative_from_returns(self._data[[column]], value, keepna)
                )
            newreturns = pd.concat(newreturnss, axis=1)

        return CumulativeReturns(
            newreturns, name=name, check_base=check_base, periods=periods
        )

    def plot_low(self, scale_x: bool = True) -> ggplot:
        """
        Plot all returns low price

        Example
        -------

        >>> from portfolio_plan.financialseries import Returns, Frequency
        >>> from portfolio_plan.resource import File
        >>> from portfolio_plan.data import example_prices_path
        >>> from portfolio_plan.utils import returns_from_prices

        >>> resource = File(path=example_prices_path(), frequency=Frequency.BDAILY)
        >>> prices = resource.fetch(
        ...     ["A", "B"],
        ...     start_date="2025-01-01",
        ...     end_date="2025-01-31"
        ... )
        >>> returns = Returns(returns_from_prices(prices.data))
        >>> returns.plot_low()

        >>> returns.cumulative().plot_low()
        """
        return plot_returns(self._data["low"], scale_x)

    def plot_high(self, scale_x: bool = True) -> ggplot:
        """
        Plot all returns high price
        """
        return plot_returns(self._data["high"], scale_x)

    def plot_open(self, scale_x: bool = True) -> ggplot:
        """
        Plot all returns open price
        """

        return plot_returns(self._data["open"], scale_x)

    def plot_close(self, scale_x: bool = True) -> ggplot:
        """
        Plot all returns close price
        """
        return plot_returns(self._data["close"], scale_x)

    def plot(self, scale_x: bool = True):
        """
        Plot all returns candle chart
        """
        raise NotImplementedError()

    @property
    def is_unique_symbol(self) -> bool:
        data = self.data
        usymbols = data.index.get_level_values("symbol").unique()
        return usymbols.shape[0] == 1

    @property
    def is_singleton(self):
        return self.data.shape[0] == 1


class Comparison:
    def __init__(self, objs: List[FinancialSeries]):
        """ """
        if not isinstance(objs, list):
            raise TypeError("Expected a list")
        if not objs:
            raise ValueError("At least one element required")
        if not isinstance(objs[0], (FinancialSeries)):
            raise TypeError("Elements must be a concrete type of FinancialSeries")
        for i in range(1, len(objs)):
            current = objs[i]
            previous = objs[i - 1]
            if not isinstance(current, type(previous)):
                raise TypeError("All elements should be of same type")

        self._objs = objs

    def _plot_price_dimension(
        self, column: Literal["open", "low", "high", "close"], scale_x: bool
    ):
        objs = self._objs
        match column:
            case "open":
                previous = objs[0].plot_open(scale_x=scale_x)
            case "close":
                previous = objs[0].plot_close(scale_x=scale_x)
            case "high":
                previous = objs[0].plot_high(scale_x=scale_x)
            case "low":
                previous = objs[0].plot_low(scale_x=scale_x)
            case _:
                raise ValueError("not one of accepted column values")
        if objs[0].name:
            text = _geom_text(data=objs[0].data[column], text=objs[0].name)
            previous += text

        if len(objs) == 1:
            return previous

        current = previous
        for i in range(1, len(objs)):
            current_obj = objs[i]
            if isinstance(current_obj, CumulativeReturns):
                lines = _geom_lines(
                    data=current_obj.data[column], periods=current_obj.periods
                )
            else:
                lines = _geom_lines(data=current_obj.data[column])
            for geom in lines:
                current += geom
            if current_obj.name:
                text = _geom_text(data=current_obj.data[column], text=current_obj.name)
                current += text
        return current

    def plot_open(self, scale_x: bool = True) -> ggplot:
        return self._plot_price_dimension("open", scale_x=scale_x)

    def plot_close(self, scale_x: bool = True) -> ggplot:
        return self._plot_price_dimension("close", scale_x=scale_x)

    def plot_low(self, scale_x: bool = True) -> ggplot:
        return self._plot_price_dimension("low", scale_x=scale_x)

    def plot_high(self, scale_x: bool = True):
        return self._plot_price_dimension("high", scale_x=scale_x)

    def plot(self, scale_x: bool = True):
        raise NotImplementedError()


def vstack(prices_list: List[Prices] | List[Returns]) -> pd.DataFrame:
    dfs = [el.data for el in prices_list]
    return pd.concat(dfs, axis=0)  # rowwise



"""
Copy pasted from
https://gist.github.com/sondalex/d4d76c44983b485c53cb777f71ca4185
"""

from dataclasses import dataclass
from typing import Any, Dict, List


class Box:
    def __init__(self, data: Any):
        super().__setattr__(
            "_mutable",
            (
                "top_connected",
                "bottom_connected",
                "data",
            ),
        )
        self.data = data
        self.top_connected = False
        self.bottom_connected = False

    def __str__(self):
        ltop = "┌"
        rtop = "┐"
        lbottom = "└"
        rbottom = "┘"
        vbar = "│"
        hbar = "─"
        lines = str(self.data).split("\n")
        max_length = max([len(line) for line in lines], default=0)
        nspace = 2
        total_width = nspace + max_length

        if self.top_connected:
            middle = total_width // 2
            topline = (
                ltop + hbar * middle + "┴" + hbar * (total_width - middle - 1) + rtop
            )
        else:
            topline = ltop + hbar * total_width + rtop

        if self.bottom_connected:
            middle = total_width // 2
            bottomline = (
                lbottom
                + hbar * middle
                + "┬"
                + hbar * (total_width - middle - 1)
                + rbottom
            )
        else:
            bottomline = lbottom + hbar * total_width + rbottom

        content = ""
        for i, (line) in enumerate(lines):
            extraspace = " " * (max_length - len(line))
            content += f"{vbar} {line} {extraspace}{vbar}"
            if i < len(lines) - 1:
                content += "\n"

        return f"{topline}\n{content}\n{bottomline}"

    def __repr__(self):
        return self.__str__()

    def __setattr__(self, name, value):
        if name == "_mutable":
            super().__setattr__(name, value)
            return

        if hasattr(self, "_mutable") and name not in self._mutable:
            raise ValueError(f"Cannot change {name}. Only {self._mutable} are settable")

        if name in self._mutable and name != "data" and not isinstance(value, bool):
            raise ValueError("Value must be a boolean")

        super().__setattr__(name, value)


@dataclass
class Position:
    x: int
    y: int
    width: int
    height: int = 3


def calculate_positions(root_box: Box, child_boxes: List[Box]) -> Dict[Box, Position]:
    positions = {}

    def get_box_width(box: Box) -> int:
        lines = str(box.data).split("\n")  # Get raw data lines
        max_line_length = max([len(line) for line in lines], default=0)
        return max_line_length + 4

    root_width = get_box_width(root_box)
    total_children_width = (
        sum(get_box_width(box) for box in child_boxes) + (len(child_boxes) - 1) * 4
    )
    start_x = max(0, (total_children_width - root_width) // 2)

    positions[root_box] = Position(x=start_x, y=0, width=root_width)

    current_x = 0
    for i, box in enumerate(child_boxes):
        box_width = get_box_width(box)
        positions[box] = Position(x=current_x, y=4, width=box_width)
        current_x += box_width + 4

        box.top_connected = True
        if i == len(child_boxes) // 2:
            root_box.bottom_connected = True

    return positions


def connect(root_box: Box, child_boxes: List[Box]) -> str:
    positions = calculate_positions(root_box, child_boxes)

    max_width = max(pos.x + pos.width for pos in positions.values())
    max_height = max(pos.y + pos.height for pos in positions.values()) + 1 + 1

    grid = [[" " for _ in range(max_width + 1)] for _ in range(max_height)]

    for box, pos in positions.items():
        box_lines = str(box).split("\n")

        y_offset = 1 if box in child_boxes else 0

        for i, line in enumerate(box_lines):
            for j, char in enumerate(line):
                if pos.y + i + y_offset < len(grid) and pos.x + j < len(grid[0]):
                    grid[pos.y + i + y_offset][pos.x + j] = char

    if root_box.bottom_connected and child_boxes:
        root_pos = positions[root_box]
        connection_x = root_pos.x + root_pos.width // 2

        grid[root_pos.y + root_pos.height][connection_x] = "│"

        grid[root_pos.y + root_pos.height + 1][connection_x] = "┴"

        first_node_pos = positions[child_boxes[0]]
        last_node_pos = positions[child_boxes[-1]]
        left_x = first_node_pos.x + (first_node_pos.width // 2)
        right_x = last_node_pos.x + (last_node_pos.width // 2)

        grid[root_pos.y + root_pos.height + 1][left_x] = "┌"
        grid[root_pos.y + root_pos.height + 1][right_x] = "┐"

        for x in range(left_x + 1, right_x):
            if x != connection_x:
                grid[root_pos.y + root_pos.height + 1][x] = "─"

        for child_box in child_boxes[1:-1]:
            child_pos = positions[child_box]
            child_x = child_pos.x + (child_pos.width // 2)
            if child_x != connection_x:
                grid[root_pos.y + root_pos.height + 1][child_x] = "┬"
                pass

    return "\n".join("".join(row).rstrip() for row in grid)


def ascii_graph(data: Dict[str, str], nodes: List[Dict[str, str]]) -> str:
    data_box = Box(data)
    nodes_boxes = [Box(node) for node in nodes]
    return connect(data_box, nodes_boxes)



from abc import ABC, abstractmethod

import pandas as pd
from plotnine.ggplot import ggplot

from portfolio_plan.utils import _create_str_stack


class FinancialSeries(ABC):
    def __init__(self, data: pd.DataFrame, name: str) -> None:
        self.validate_data(data)
        self._data = data
        self._name = name

    @abstractmethod
    def validate_data(self, data: pd.DataFrame):
        pass

    def __str__(self) -> str:
        dates = self._data.index.get_level_values("date")
        symbols = self._data.index.get_level_values("symbol")
        min_date = dates.min()
        max_date = dates.max()
        name = self.__class__.__name__
        stack: str = _create_str_stack(
            f"{name} for period {min_date} to {max_date}", symbols.unique().tolist()
        )
        return stack

    @property
    def name(self) -> str:
        return self._name

    def __repr__(self) -> str:
        return self.__str__()

    @abstractmethod
    def plot_open(self, scale_x: bool = True) -> ggplot:
        """
        Plot open price
        """
        pass

    @abstractmethod
    def plot_close(self, scale_x: bool = True) -> ggplot:
        """
        Plot close price
        """
        pass

    @abstractmethod
    def plot_low(self, scale_x: bool = True) -> ggplot:
        """
        Plot low price
        """
        pass

    @abstractmethod
    def plot_high(self, scale_x: bool = True) -> ggplot:
        """
        Plot high price
        """
        pass

    @abstractmethod
    def plot(self, scale_x: bool = True) -> ggplot:
        """
        Plot OHLC price
        """
        pass

    @property
    def data(self) -> pd.DataFrame:
        return self._data



from importlib import resources
from pathlib import PosixPath


def example_prices_path() -> PosixPath:
    filename = "mock_example.parquet"
    files = resources.files("portfolio_plan.data")
    for file in files.iterdir():
        if file.name == filename:
            if not isinstance(file, PosixPath):
                raise RuntimeError("Uncaught error, can you raise an issue ?")
            return file
    raise FileNotFoundError(f"{filename} file not found in resources")


def mock_prices_path() -> PosixPath:
    filename = "mock.parquet"
    files = resources.files("portfolio_plan.data")
    for file in files.iterdir():
        if file.name == filename:
            if not isinstance(file, PosixPath):
                raise RuntimeError("Uncaught error, can you raise an issue ?")
            return file
    raise FileNotFoundError(f"{filename} file not found in resources")


def mock_returns_path() -> PosixPath:
    filename = "mock_returns.parquet"
    files = resources.files("portfolio_plan.data")
    for file in files.iterdir():
        if file.name == filename:
            if not isinstance(file, PosixPath):
                raise RuntimeError("Uncaught error, can you raise an issue ?")
            return file
    raise FileNotFoundError(f"{filename} file not found in resources")


def mock_cumulative_returns_path() -> PosixPath:
    filename = "mock_cumulative_returns.parquet"
    files = resources.files("portfolio_plan.data")
    for file in files.iterdir():
        if file.name == filename:
            if not isinstance(file, PosixPath):
                raise RuntimeError("Uncaught error, can you raise an issue ?")
            return file
    raise FileNotFoundError(f"{filename} file not found in resources")



import numpy as np
import pandas as pd
import pytest

from portfolio_plan import const
from portfolio_plan.errors import ValidationError
from portfolio_plan.financialseries import Returns


class TestReturns:
    @pytest.fixture
    def sample_returns_data(self):
        """
        Create sample returns data for testing
        """
        values = [
            np.nan,  # A, 2020-01-01
            ((110.0 - 100.0) / 100.0),  # A, 2020-01-02
            ((115.0 - 110.0) / 110.0),  # A, 2020-01-03
            ((120.0 - 115.0) / 115.0),  # A, 2020-01-04
            np.nan,  # B, 2020-01-01
            ((210.0 - 200.0) / 200.0),  # B, 2020-01-02
            ((220.0 - 210.0) / 210.0),  # B, 2020-01-03
            ((225.0 - 220.0) / 220.0),  # B, 2020-01-04
        ]

        index_tuples = [
            ("A", pd.Timestamp("2020-01-01")),
            ("A", pd.Timestamp("2020-01-02")),
            ("A", pd.Timestamp("2020-01-03")),
            ("A", pd.Timestamp("2020-01-04")),
            ("B", pd.Timestamp("2020-01-01")),
            ("B", pd.Timestamp("2020-01-02")),
            ("B", pd.Timestamp("2020-01-03")),
            ("B", pd.Timestamp("2020-01-04")),
        ]

        return pd.DataFrame(
            {
                "open": values,
                "close": values,
                "high": values,
                "low": values,
            },
            index=pd.MultiIndex.from_tuples(
                index_tuples,
                names=["symbol", "date"],
            ),
        )

    def test_returns_initialization(self, sample_returns_data):
        """Test that Returns class initializes correctly with valid data"""
        returns = Returns(sample_returns_data, name="Test Returns")
        assert returns.name == "Test Returns"
        pd.testing.assert_frame_equal(returns.data, sample_returns_data)

    def test_returns_initialization_invalid_data(self):
        invalid_values = [
            0.1,  # A, 2020-01-01 (not NaN)
            ((110.0 - 100.0) / 100.0),  # A, 2020-01-02
            ((115.0 - 110.0) / 110.0),  # A, 2020-01-03
            ((120.0 - 115.0) / 115.0),  # A, 2020-01-04
            np.nan,  # B, 2020-01-01 (correctly NaN)
            ((210.0 - 200.0) / 200.0),  # B, 2020-01-02
            ((220.0 - 210.0) / 210.0),  # B, 2020-01-03
            ((225.0 - 220.0) / 220.0),  # B, 2020-01-04
        ]

        invalid_index = pd.MultiIndex.from_tuples(
            [
                ("A", pd.Timestamp("2020-01-01")),
                ("A", pd.Timestamp("2020-01-02")),
                ("A", pd.Timestamp("2020-01-03")),
                ("A", pd.Timestamp("2020-01-04")),
                ("B", pd.Timestamp("2020-01-01")),
                ("B", pd.Timestamp("2020-01-02")),
                ("B", pd.Timestamp("2020-01-03")),
                ("B", pd.Timestamp("2020-01-04")),
            ],
            names=["symbol", "date"],
        )

        invalid_data = pd.DataFrame(
            {
                "open": invalid_values,
                "close": invalid_values,
                "high": invalid_values,
                "low": invalid_values,
            },
            index=invalid_index,
        )

        with pytest.raises(ValidationError):
            Returns(invalid_data, name="Invalid Returns")

        valid_values = [
            np.nan,  # A, 2020-01-01 (correctly NaN)
            ((110.0 - 100.0) / 100.0),  # A, 2020-01-02
            ((115.0 - 110.0) / 110.0),  # A, 2020-01-03
            ((120.0 - 115.0) / 115.0),  # A, 2020-01-04
            np.nan,  # B, 2020-01-01 (correctly NaN)
            ((210.0 - 200.0) / 200.0),  # B, 2020-01-02
            ((220.0 - 210.0) / 210.0),  # B, 2020-01-03
            ((225.0 - 220.0) / 220.0),  # B, 2020-01-04
        ]

        valid_data = pd.DataFrame(
            {
                "open": valid_values,
                "close": valid_values,
                "high": valid_values,
                "low": valid_values,
            },
            index=invalid_index,
        )

        Returns(valid_data, name="Valid Returns")

    def test_avg_without_weights(self, sample_returns_data):
        """Test the avg method without specifying weights"""
        returns = Returns(sample_returns_data, name="Test Returns")
        avg_returns = returns.avg(weights={})

        # Calculate expected results manually: average across symbols for each date
        expected_values = [
            np.nan,  # 2020-01-01 (all NaN)
            (((110.0 - 100.0) / 100.0) + ((210.0 - 200.0) / 200.0)) / 2,  # 2020-01-02
            (((115.0 - 110.0) / 110.0) + ((220.0 - 210.0) / 210.0)) / 2,  # 2020-01-03
            (((120.0 - 115.0) / 115.0) + ((225.0 - 220.0) / 220.0)) / 2,  # 2020-01-04
        ]

        expected_data = pd.DataFrame(
            {
                "open": expected_values,
                "close": expected_values,
                "high": expected_values,
                "low": expected_values,
            },
            index=pd.MultiIndex.from_tuples(
                [
                    (const.DEFAULT_PORTFOLIO_NAME, pd.Timestamp("2020-01-01")),
                    (const.DEFAULT_PORTFOLIO_NAME, pd.Timestamp("2020-01-02")),
                    (const.DEFAULT_PORTFOLIO_NAME, pd.Timestamp("2020-01-03")),
                    (const.DEFAULT_PORTFOLIO_NAME, pd.Timestamp("2020-01-04")),
                ],
                names=["symbol", "date"],
            ),
        )

        pd.testing.assert_frame_equal(avg_returns.data, expected_data)

    def test_avg_with_weights(self, sample_returns_data):
        """Test the avg method with specified weights"""
        returns = Returns(sample_returns_data, name="Test Returns")

        weights = {"A": 0.6, "B": 0.4}

        avg_returns = returns.avg(weights=weights, portfolio_name="Weighted Portfolio")

        # With the fixed implementation, expected values should be a proper weighted average
        expected_values = [
            np.nan,  # 2020-01-01
            ((110.0 - 100.0) / 100.0) * 0.6
            + ((210.0 - 200.0) / 200.0) * 0.4,  # 2020-01-02
            ((115.0 - 110.0) / 110.0) * 0.6
            + ((220.0 - 210.0) / 210.0) * 0.4,  # 2020-01-03
            ((120.0 - 115.0) / 115.0) * 0.6
            + ((225.0 - 220.0) / 220.0) * 0.4,  # 2020-01-04
        ]

        expected_data = pd.DataFrame(
            {
                "open": expected_values,
                "close": expected_values,
                "high": expected_values,
                "low": expected_values,
            },
            index=pd.MultiIndex.from_tuples(
                [
                    ("Weighted Portfolio", pd.Timestamp("2020-01-01")),
                    ("Weighted Portfolio", pd.Timestamp("2020-01-02")),
                    ("Weighted Portfolio", pd.Timestamp("2020-01-03")),
                    ("Weighted Portfolio", pd.Timestamp("2020-01-04")),
                ],
                names=["symbol", "date"],
            ),
        )

        pd.testing.assert_frame_equal(avg_returns.data, expected_data)



from typing import Dict, Literal, TypedDict

from plotnine.geoms import geom_line, geom_text
from plotnine.scales.scale_color import scale_fill_brewer
from plotnine.scales.scale_manual import scale_color_manual, scale_fill_manual
from plotnine.themes.elements import element_line, element_rect, element_text
from plotnine.themes.theme import theme
from plotnine.themes.theme_bw import theme_bw

THEME_ROSE_PINE_BASE_SIZE = 11


class ColorPalette(TypedDict):
    love: str
    gold: str
    rose: str
    pine: str
    foam: str
    iris: str


class Palette(TypedDict):
    background: str
    surface: str
    text: str
    highlight_major: str
    highlight_minor: str
    border: str
    color: ColorPalette


class theme_rose_pine(theme_bw):
    def __init__(
        self,
        variant: Literal["main", "moon", "dawn"] = "main",
        geom_color: Literal["love", "gold", "rose", "pine", "foam", "iris"] = "love",
        base_size: int = THEME_ROSE_PINE_BASE_SIZE,
        base_family=None,
    ):
        """
        Initialize the Rose Pine theme with support for Main, Moon, and Dawn variants.

        Parameters:
        -----------
        variant: str
            "main", "moon", or "dawn" to select the color palette.
        color: str
            Default geom color category.
            See category at `rosé pine https://rosepinetheme.com/palette/`
        base_size: int
            Base font size.
        base_family: str
            Base font family.
        """
        super().__init__(base_size, base_family)

        palettes: Dict[str, Palette] = {
            "main": {
                "background": "#191724",
                "surface": "#1f1d2e",
                "text": "#e0def4",
                "highlight_major": "#eb6f92",
                "highlight_minor": "#f6c177",
                "border": "#6e6a86",
                "color": {
                    "love": "#eb6f92",
                    "gold": "#f6c177",
                    "rose": "#ebbcba",
                    "pine": "#31748f",
                    "foam": "#9ccfd8",
                    "iris": "#c4a7e7",
                },
            },
            "moon": {
                "background": "#232136",
                "surface": "#2a273f",
                "text": "#e0def4",
                "highlight_major": "#eb6f92",
                "highlight_minor": "#f6c177",
                "border": "#6e6a86",
                "color": {
                    "love": "#eb6f92",
                    "gold": "#f6c177",
                    "rose": "#ea9a97",
                    "pine": "#3e8fb0",
                    "foam": "#9ccfd8",
                    "iris": "#c4a7e7",
                },
            },
            "dawn": {
                "background": "#faf4ed",
                "surface": "#fffaf3",
                "text": "#575279",
                "highlight_major": "#b4637a",
                "highlight_minor": "#ea9d34",
                "border": "#9893a5",
                "color": {
                    "love": "#b4637a",
                    "gold": "#ea9d34",
                    "rose": "#d7827e",
                    "pine": "#286983",
                    "foam": "#56949f",
                    "iris": "#907aa9",
                },
            },
        }

        if variant not in palettes:
            raise ValueError(
                f"Invalid variant '{variant}'. Choose from 'main', 'moon', or 'dawn'."
            )

        palette = palettes[variant]
        geom_text.DEFAULT_AES["color"] = palette["text"]
        geom_line.DEFAULT_AES["color"] = palette["color"][geom_color]
        self += theme(  # type: ignore
            text=element_text(color=palette["text"], face="plain"),
            plot_background=element_rect(
                fill=palette["background"], color=palette["border"]
            ),
            plot_title=element_text(color=palette["text"]),
            axis_title=element_text(color=palette["text"]),
            panel_background=element_rect(
                fill=palette["surface"], color=palette["border"]
            ),
            panel_grid_major=element_line(
                color=palette["highlight_major"], size=0.5, linetype="dotted"
            ),
            panel_grid_minor=element_line(
                color=palette["highlight_minor"], size=0.25, linetype="dotted"
            ),
            axis_text=element_text(color=palette["text"]),
            legend_text=element_text(color=palette["text"]),
            legend_title=element_text(color=palette["text"]),
            legend_background=element_rect(fill=palette["background"]),
            legend_key=element_rect(fill=palette["background"], color="none"),
            strip_background=element_rect(
                fill=palette["surface"], color=palette["border"]
            ),
            strip_text_x=element_text(color=palette["text"]),
            strip_text_y=element_text(color=palette["text"], angle=-90),
        )


class scale_rose_pine_discrete(scale_color_manual):
    def __init__(self, variant="main"):
        """
        Custom discrete color scale for the Rose Pine theme.

        Parameters:
        - variant (str): "main", "moon", or "dawn" to select the color palette.
        """
        # Define color palettes for each variant
        palettes = {
            "main": {
                "love": "#eb6f92",
                "gold": "#f6c177",
                "rose": "#ebbcba",
                "pine": "#31748f",
                "foam": "#9ccfd8",
                "iris": "#c4a7e7",
            },
            "moon": {
                "love": "#eb6f92",
                "gold": "#f6c177",
                "rose": "#ea9a97",
                "pine": "#3e8fb0",
                "foam": "#9ccfd8",
                "iris": "#c4a7e7",
            },
            "dawn": {
                "love": "#b4637a",
                "gold": "#ea9d34",
                "rose": "#d7827e",
                "pine": "#286983",
                "foam": "#56949f",
                "iris": "#907aa9",
            },
        }

        if variant not in palettes:
            raise ValueError(
                f"Invalid variant '{variant}'. Choose from 'main', 'moon', or 'dawn'."
            )

        palette = palettes[variant]

        super().__init__(values=list(palette.values()))


class scale_rose_pine_fill_discrete(scale_fill_manual):
    def __init__(self, variant="main"):
        palettes = {
            "main": {
                "love": "#eb6f92",
                "gold": "#f6c177",
                "rose": "#ebbcba",
                "pine": "#31748f",
                "foam": "#9ccfd8",
                "iris": "#c4a7e7",
            },
            "moon": {
                "love": "#eb6f92",
                "gold": "#f6c177",
                "rose": "#ea9a97",
                "pine": "#3e8fb0",
                "foam": "#9ccfd8",
                "iris": "#c4a7e7",
            },
            "dawn": {
                "love": "#b4637a",
                "gold": "#ea9d34",
                "rose": "#d7827e",
                "pine": "#286983",
                "foam": "#56949f",
                "iris": "#907aa9",
            },
        }

        if variant not in palettes:
            raise ValueError(
                f"Invalid variant '{variant}'. Choose from 'main', 'moon', or 'dawn'."
            )

        palette = palettes[variant]
        sorted_values = map(lambda key: palette[key], sorted(palette.keys()))
        super().__init__(values=list(sorted_values))


class scale_brewer_fill_discrete(scale_fill_brewer):
    # TODO: test visibility with variant main, moon, dawn
    def __init__(self, palette="Set3"):
        """
        Parameters
        ----------
        palette: str
            Refer to `ggplot2 https://ggplot2.tidyverse.org/reference/scale_brewer.html`
            and to `ggplot2 book https://ggplot2-book.org/scales-colour#brewer-scales`
        """
        super().__init__(type="qual", palette=palette)


class scale_alphabet_fill_discrete(scale_fill_manual):
    """
    Based on `pals https://kwstat.github.io/pals/` alphabet
    """

    def __init__(self):
        """ """
        values = [
            "#F0A0FF",
            "#0075DC",
            "#993F00",
            "#4C005C",
            "#191919",
            "#005C31",
            "#2BCE48",
            "#FFCC99",
            "#808080",
            "#94FFB5",
            "#8F7C00",
            "#9DCC00",
            "#C20088",
            "#003380",
            "#FFA405",
            "#FFA8BB",
            "#426600",
            "#FF0010",
            "#5EF1F2",
            "#00998F",
            "#E0FF66",
            "#740AFF",
            "#990000",
            "#FFFF80",
            "#FFE100",
            "#FF5005",
        ]
        super().__init__(values=values)



from portfolio_plan._properties import Frequency, Weights
from portfolio_plan.core import Asset, Assets, JoinedPlan, Plan, Portfolio

__all__ = [
    "Frequency",
    "Weights",
    "Asset",
    "Assets",
    "Plan",
    "JoinedPlan",
    "Portfolio",
]
